from pymongo import DESCENDING

from CveXplore.common.cpe_converters import create_cpe_regex_string
from CveXplore.common.data_source_connection import DatasourceConnection


class Cpe(DatasourceConnection):
    """
    Cpe database object
    """

    def __init__(self, **kwargs):
        super().__init__("cpe")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def iter_cves_matching_cpe(self, vuln_prod_search: bool = False):
        """
        Generator function for iterating over cve's matching this CPE. By default the search will be made matching
        the configuration fields of the cves documents.
        """

        cpe_searchField = (
            "vulnerable_product" if vuln_prod_search else "vulnerable_configuration"
        )

        cpe_regex_string = create_cpe_regex_string(self.cpeName)

        results = self.datasource_connection.store_cves.find(
            {cpe_searchField: {"$regex": cpe_regex_string}}
        ).sort("cvss", DESCENDING)

        for each in results:
            if each is not None:
                yield each
            else:
                yield None

    def to_cve_summary(self, vuln_prod_search: bool = False) -> dict:
        """
        Method to request all cve's from the database based on this cpe object
        """
        all_cves = list(self.iter_cves_matching_cpe(vuln_prod_search=vuln_prod_search))
        data_cves = [
            d.to_dict(
                "id",
                "cvss",
                "cvss3",
                "published",
                "modified",
                "summary",
                "references",
                "status",
            )
            for d in all_cves
        ]
        data_cpe = self.to_dict()
        data_cpe["cvecount"] = len(data_cves)
        data_cpe["cves"] = data_cves
        return data_cpe

    def __repr__(self):
        """String representation of object"""
        return f"<< Cpe:{self.id} >>"



import pymongo

from CveXplore.core.api_base_class import ApiBaseClass


class CveSearchApi(ApiBaseClass):
    """
    The CveSearchApi handles the different arguments in order to perform a query to a specific Cve Search api endpoint.
    It mimics the pymongo cursor's behaviour to provide an ambiguous way to talk to either an API or a mongodb.
    """

    def __init__(
        self,
        db_collection,
        filter: dict = None,
        limit: int = None,
        skip: int = None,
        sort: tuple = None,
    ):
        """
        Create a new CveSearchApi object.
        """
        super().__init__(
            address=db_collection.baseurl,
            api_path=db_collection.api_path,
            proxies=db_collection.proxies,
            user_agent=db_collection.user_agent,
        )
        from CveXplore.common.db_obj_mapping import database_objects_mapping

        self.database_objects_mapping = database_objects_mapping

        self.db_collection = db_collection

        self.collname = db_collection.collname

        data = filter
        if data is None:
            data = {}

        self.__data = {"retrieve": self.collname, "dict_filter": data}

        self.__empty = False

        self.__limit = limit
        self.__skip = skip
        self.__sort = sort

        self.data_queue = None

    def __repr__(self):
        """return a string representation of the obj GenericApi"""
        return f"<<CveSearchApi:({self.db_collection.address[0]}, {self.db_collection.address[1]})>>"

    def query(self):
        """
        Endpoint for free query to cve search data
        """

        results = self.call(method="POST", resource="query", data=self.__data)

        if isinstance(results, str):
            self.data_queue = None
            return

        try:
            if len(results["data"]) > 0:
                self.data_queue = results["data"]
        except Exception:
            self.data_queue = results

    def limit(self, value: int):
        """
        Method to limit the amount of returned data
        """

        if not isinstance(value, int):
            raise TypeError("limit must be an integer")

        self.__limit = value

        self.__data["limit"] = self.__limit

        return self

    def skip(self, value: int):
        """
        Method to skip the given amount of records before returning the data
        """

        if not isinstance(value, int):
            raise TypeError("skip must be an integer")

        self.__skip = value

        self.__data["skip"] = self.__skip

        return self

    def sort(self, field: str, direction: int):
        """
        Method to sort the returned data
        """

        if direction == pymongo.DESCENDING:
            direction = "DESC"
        else:
            direction = "ASC"

        self.__sort = (field, direction)

        self.__data["sort"] = self.__sort[0]
        self.__data["sort_dir"] = self.__sort[1]

        return self

    def __iter__(self):
        """Make this class an iterator"""
        self.query()
        return self

    def next(self):
        """Iterate to the results and return database objects"""
        if self.__empty:
            raise StopIteration
        if self.data_queue is None:
            raise StopIteration
        try:
            if len(self.data_queue):
                return self.database_objects_mapping[self.collname](
                    **self.data_queue.pop()
                )
            else:
                raise StopIteration
        except TypeError:
            # We've received a Response object
            raise StopIteration

    __next__ = next



class ApiException(Exception):
    pass


class ApiErrorException(ApiException):
    pass


class ApiDataError(ApiException):
    pass


class ApiDataRetrievalFailed(ApiException):
    pass


class ApiMaxRetryError(ApiException):
    pass



import click
import pymongo

from CveXplore.cli_cmds.cli_utils.utils import printer
from CveXplore.cli_cmds.mutex_options.mutex import Mutex


@click.group(
    "find",
    invoke_without_command=True,
    help="Perform find queries on a single collection",
)
@click.option(
    "-c",
    "--collection",
    required=True,
    type=click.Choice(["capec", "cpe", "cwe", "via4", "cves"], case_sensitive=False),
    help="Collection to query",
)
@click.option("-f", "--field", required=True, help="Field to query")
@click.option("-v", "--value", required=True, help="Value to query")
@click.option(
    "-m",
    "--match",
    is_flag=True,
    help="Use match for searching (default)",
    cls=Mutex,
    not_required_if=["regex"],
)
@click.option(
    "-r",
    "--regex",
    is_flag=True,
    help="Use regex for searching",
    cls=Mutex,
    not_required_if=["match"],
)
@click.option("-l", "--limit", default=10, help="Query limit")
@click.option(
    "-lf",
    "--limit-field",
    help="Limit the return the this field(s) (could be multiple)",
    multiple=True,
)
@click.option(
    "-s", "--sort", is_flag=True, help="Sort DESCENDING (for match and regex only)"
)
@click.option(
    "-o",
    "--output",
    default="json",
    help="Set the desired output format (defaults to json)",
    type=click.Choice(["json", "csv", "xml", "html"], case_sensitive=False),
)
@click.pass_context
def find_cmd(
    ctx, collection, field, value, match, regex, limit, limit_field, sort, output
):
    if not sort:
        sorting = pymongo.ASCENDING
    else:
        sorting = pymongo.DESCENDING
    try:
        if regex:
            ret_list = (
                getattr(getattr(ctx.obj["data_source"], collection), field)
                .search(value)
                .limit(limit)
                .sort(field, sorting)
            )
        elif match:
            ret_list = (
                getattr(getattr(ctx.obj["data_source"], collection), field)
                .find(value)
                .limit(limit)
                .sort(field, sorting)
            )
        else:
            ret_list = ctx.obj["data_source"].get_single_store_entries(
                (collection, {field: value}), limit=limit
            )
    except AttributeError:
        click.echo(
            f"Field: {field} is not mapped for the collection: {collection}; you can choose "
            f"from: {getattr(ctx.obj['data_source'], collection).mapped_fields(collection=collection)}"
        )
        return

    print(len(limit_field))

    try:
        if len(limit_field) != 0:
            result = [result.to_dict(*limit_field, field) for result in ret_list]
        else:
            result = [result.to_dict() for result in ret_list]
    except TypeError:
        result = []

    if ctx.invoked_subcommand is None:
        printer(input_data=result, output=output)
    else:
        ctx.obj["RESULT"] = result



import re
from functools import reduce
from typing import List

from pymongo import DESCENDING

from CveXplore.database.helpers.generic_db import GenericDatabaseFactory
from CveXplore.errors.validation import CveNumberValidationError
from CveXplore.objects.cvexplore_object import CveXploreObject


class CvesDatabaseFunctions(GenericDatabaseFactory):
    """
    The CvesDatabaseFunctions is a specific class that provides the cves attribute of a CveXplore instance additional
    functions that only apply to the 'cves' collection
    """

    def __init__(self, collection: str):
        super().__init__(collection)

    def get_cves_for_vendor(
        self, vendor: str, limit: int = 0
    ) -> List[CveXploreObject] | None:
        """
        Function to return cves based on a given vendor. By default, to result is sorted descending on th cvss field.
        """

        the_result = list(
            self.datasource_collection_connection.find({"vendors": vendor})
            .limit(limit)
            .sort("cvss", DESCENDING)
        )

        if len(the_result) != 0:
            return the_result
        else:
            return None

    def __parse_cve_id(self, doc_id: str) -> str:
        # first try to match full cve number format
        reg_match = re.compile(r"[cC][vV][eE]-\d{4}-\d{4,10}")
        if reg_match.match(doc_id) is not None:
            doc_id = doc_id.upper()
        else:
            part_match = re.compile(r"\d{4}-\d{4,10}")
            if part_match.match(doc_id) is not None:
                doc_id = f"CVE-{doc_id}"
            else:
                raise CveNumberValidationError(
                    "Could not validate the CVE number. The number format should be either "
                    "CVE-2000-0001, cve-2000-0001 or 2000-0001."
                )
        return doc_id

    def get_by_id(self, doc_id: str):
        """
        Method to retrieve a single CVE from the database by its CVE ID number.
        The number format should be either CVE-2000-0001, cve-2000-0001 or 2000-0001.
        """
        doc_id = self.__parse_cve_id(doc_id)

        if not isinstance(doc_id, str):
            try:
                doc_id = str(doc_id)
            except ValueError:
                return "Provided value is not a string nor can it be cast to one"

        return self.datasource_collection_connection.find_one({"id": doc_id})

    def _field_list(self, doc_id: str) -> list:
        """
        Method to fetch all field names from a specific collection
        """
        doc_id = self.__parse_cve_id(doc_id)

        return sorted(
            list(
                reduce(
                    lambda all_keys, rec_keys: all_keys | set(rec_keys),
                    map(
                        lambda d: d.to_dict(),
                        [
                            self.datasource_collection_connection.find_one(
                                {"id": doc_id}
                            )
                        ],
                    ),
                    set(),
                )
            )
        )


class CpeDatabaseFunctions(GenericDatabaseFactory):
    """
    The CpeDatabaseFunctions is a specific class that provides the cpe attribute of a CveXplore instance additional
    functions that only apply to the 'cpe' collection
    """

    def __init__(self, collection: str):
        super().__init__(collection)

    def search_active_cpes(
        self, field: str, value: str, limit: int = 0, sorting: int = 1
    ) -> List[CveXploreObject] | None:
        """
        Function to regex search for cpe based on value in string. Only active (deprecated == false) cpe records are
        returned.
        """
        regex = re.compile(value, re.IGNORECASE)

        query = {"$and": [{field: {"$regex": regex}}, {"deprecated": False}]}

        the_result = list(
            self.datasource_collection_connection.find(query)
            .limit(limit)
            .sort(field, sorting)
        )

        if len(the_result) != 0:
            return the_result
        else:
            return None

    def find_active_cpes(
        self, field: str, value: str, limit: int = 0, sorting: int = 1
    ) -> List[CveXploreObject] | None:
        """
        Function to find cpe based on value in string. Only active (deprecated == false) cpe records are
        returned.
        """
        query = {"$and": [{field: value}, {"deprecated": False}]}

        the_result = list(
            self.datasource_collection_connection.find(query)
            .limit(limit)
            .sort(field, sorting)
        )

        if len(the_result) != 0:
            return the_result
        else:
            return None


class CapecDatabaseFunctions(GenericDatabaseFactory):
    """
    The CapecDatabaseFunctions is a specific class that provides the capec attribute of a CveXplore instance additional
    functions that only apply to the 'capec' collection
    """

    def __init__(self, collection: str):
        super().__init__(collection)


class CWEDatabaseFunctions(GenericDatabaseFactory):
    """
    The CWEDatabaseFunctions is a specific class that provides the cwe attribute of a CveXplore instance additional
    functions that only apply to the 'cwe' collection
    """

    def __init__(self, collection: str):
        super().__init__(collection)



import datetime
import glob
import hashlib
import json
import shutil
import time
from typing import Any, Tuple
from xml.sax import make_parser

from dateutil.parser import parse as parse_datetime
from tqdm import tqdm

from CveXplore.core.database_actions.db_action import DatabaseAction
from CveXplore.core.database_maintenance.api_handlers import NVDApiHandler
from CveXplore.core.database_maintenance.content_handlers import (
    CapecHandler,
    CWEHandler,
)
from CveXplore.core.database_maintenance.file_handlers import (
    XMLFileHandler,
    JSONFileHandler,
    CSVFileHandler,
)
from CveXplore.errors.apis import ApiDataRetrievalFailed, ApiMaxRetryError

date = datetime.datetime.now()
year = date.year + 1

# default config
defaultvalue = {"cwe": "Unknown"}


class CPEDownloads(NVDApiHandler):
    """
    Class processing CPE source files
    """

    def __init__(self):
        self.feed_type = "CPE"

        super().__init__(feed_type=self.feed_type, logger_name=__name__)

    def file_to_queue(self, *args):
        pass

    def parse_cpe_version(self, cpename: str):
        cpe_list = self.split_cpe_name(cpename)
        version_stem = cpe_list[5]

        if cpe_list[6] != "*" and cpe_list[6] != "-":
            return f"{version_stem}.{cpe_list[6]}"
        else:
            return version_stem

    def process_the_item(self, item: dict = None):
        if item is None:
            return None

        item = item["cpe"]

        # filter out deprecated CPE's if CPE_FILTER_DEPRECATED is set to True
        if self.config.CPE_FILTER_DEPRECATED:
            if item["deprecated"]:
                return None

        if "cpeName" not in item:
            return None

        title = None

        if "titles" in item:
            for t in item["titles"]:
                if t["lang"] == "en":
                    title = t["title"]

        version = self.parse_cpe_version(cpename=item["cpeName"])

        split_cpe_name = self.split_cpe_name(item["cpeName"])
        cpe = {
            "title": title,
            "cpeName": item["cpeName"],
            "vendor": split_cpe_name[3],
            "product": split_cpe_name[4],
            "version": version,
            "padded_version": self.padded_version(version),
            "stem": self.stem(item["cpeName"]),
            "cpeNameId": item["cpeNameId"],
            "lastModified": parse_datetime(item["lastModified"], ignoretz=True),
            "created": parse_datetime(item["created"], ignoretz=True),
            "deprecated": item["deprecated"],
            "deprecatedBy": item["deprecatedBy"] if item["deprecated"] else "",
        }

        sha1_hash = hashlib.sha1(
            cpe["cpeName"].encode("utf-8") + split_cpe_name[5].encode("utf-8")
        ).hexdigest()

        cpe["id"] = sha1_hash

        return cpe

    def process_downloads(self, sites: list | None = None):
        """
        Method to download and process files
        """

        self.logger.info("Starting download...")

        start_time = time.time()

        self.last_modified = datetime.datetime.now()

        self.logger.debug(
            f"do_process = {self.do_process}; is_update = {self.is_update}"
        )

        if self.do_process:
            if not self.is_update:
                try:
                    total_results = self.api_handler.get_count(
                        self.api_handler.datasource.CPE
                    )
                except ApiMaxRetryError:
                    # failed to get the count; set total_results to 0 and continue
                    total_results = 0

                self.logger.info(f"Preparing to download {total_results} CPE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(data_type="cpe"):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                if not isinstance(data_list, Exception):
                                    processed_items = [
                                        self.process_item(item)
                                        for item in data_list["products"]
                                    ]
                                    self._db_bulk_writer(
                                        processed_items, initialization_run=True
                                    )
                                    pbar.update(len(data_list["products"]))
                                else:
                                    self.logger.error(
                                        f"Retrieval of api data resulted in an Exception: {data_list}..."
                                    )
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )
            else:
                last_mod_start_date = self.database[self.feed_type.lower()].find_one(
                    {}, {"lastModified": 1}, sort=[("lastModified", -1)]
                )

                if last_mod_start_date is not None:
                    if "lastModified" in last_mod_start_date:
                        last_mod_start_date = last_mod_start_date[
                            "lastModified"
                        ] + datetime.timedelta(
                            0, 1
                        )  # add one second to prevent false results...
                    else:
                        raise KeyError(
                            "Missing field 'lastModified' from database query..."
                        )
                else:
                    self.logger.warning(
                        "No records found in the mongodb cpe collection.."
                    )
                    return

                    # Get datetime from runtime
                last_mod_end_date = datetime.datetime.now()

                try:
                    total_results = self.api_handler.get_count(
                        self.api_handler.datasource.CPE,
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    )
                except ApiMaxRetryError:
                    # failed to get the count; set total_results to 0 and continue
                    total_results = 0

                self.logger.info(f"Preparing to download {total_results} CPE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(
                        data_type="cpe",
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    ):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                if not isinstance(data_list, Exception):
                                    processed_items = [
                                        self.process_item(item)
                                        for item in data_list["products"]
                                    ]
                                    self._db_bulk_writer(processed_items)
                                    pbar.update(len(data_list["products"]))
                                else:
                                    self.logger.error(
                                        f"Retrieval of api data resulted in an Exception: {data_list}..."
                                    )
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )

            # Set the last update time in the info collection
            self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(
            f"Duration: {datetime.timedelta(seconds=time.time() - start_time)}"
        )

    def update(self, **kwargs):
        self.logger.info("CPE database update started")

        self.process_downloads()

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.database_indexer.create_indexes(collection=self.feed_type.lower())
            self.is_update = False

        self.logger.info("Finished CPE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.logger.info("CPE Database population started")

        self.is_update = False

        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        self.process_downloads()

        self.database_indexer.create_indexes(collection=self.feed_type.lower())

        self.logger.info("Finished CPE database population")

        return self.last_modified


class CVEDownloads(NVDApiHandler):
    """
    Class processing CVE source files
    """

    def __init__(self):
        self.feed_type = "CVES"

        super().__init__(feed_type=self.feed_type, logger_name=__name__)

    def get_cpe_info(self, cpeuri: str):
        query = {}

        if "versionStartExcluding" in cpeuri:
            if "versionEndExcluding" in cpeuri:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gt": self.padded_version(cpeuri["versionStartExcluding"]),
                        "$lt": self.padded_version(cpeuri["versionEndExcluding"]),
                    },
                }
            elif "versionEndIncluding" in cpeuri:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gt": self.padded_version(cpeuri["versionStartExcluding"]),
                        "$lte": self.padded_version(cpeuri["versionEndIncluding"]),
                    },
                }
            else:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gt": self.padded_version(cpeuri["versionStartExcluding"])
                    },
                }

        elif "versionStartIncluding" in cpeuri:
            if "versionEndExcluding" in cpeuri:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gte": self.padded_version(cpeuri["versionStartIncluding"]),
                        "$lt": self.padded_version(cpeuri["versionEndExcluding"]),
                    },
                }
            elif "versionEndIncluding" in cpeuri:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gte": self.padded_version(cpeuri["versionStartIncluding"]),
                        "$lte": self.padded_version(cpeuri["versionEndIncluding"]),
                    },
                }
            else:
                query = {
                    "deprecated": False,
                    "stem": self.stem(cpeuri["criteria"]),
                    "padded_version": {
                        "$gte": self.padded_version(cpeuri["versionStartIncluding"])
                    },
                }

        elif "versionEndExcluding" in cpeuri:
            query = {
                "deprecated": False,
                "stem": self.stem(cpeuri["criteria"]),
                "padded_version": {
                    "$lt": self.padded_version(cpeuri["versionEndExcluding"])
                },
            }

        elif "versionEndIncluding" in cpeuri:
            query = {
                "deprecated": False,
                "stem": self.stem(cpeuri["criteria"]),
                "padded_version": {
                    "$lte": self.padded_version(cpeuri["versionEndIncluding"])
                },
            }

        return query

    @staticmethod
    def add_if_missing(cve: dict, key: str, value: Any):
        if value not in cve[key]:
            cve[key].append(value)
        return cve

    def get_vendor_product(self, cpeUri: str):
        split_cpe_uri = self.split_cpe_name(cpeUri)
        vendor = split_cpe_uri[3]
        product = split_cpe_uri[4]
        return vendor, product

    def file_to_queue(self, *args):
        pass

    def process_the_item(self, item: dict = None):
        if item is None:
            return None

        cve = {
            "id": item["cve"]["id"],
            "assigner": item["cve"]["sourceIdentifier"],
            "status": item["cve"]["vulnStatus"],
            "published": parse_datetime(item["cve"]["published"], ignoretz=True),
            "modified": parse_datetime(item["cve"]["lastModified"], ignoretz=True),
            "lastModified": parse_datetime(item["cve"]["lastModified"], ignoretz=True),
        }

        for description in item["cve"]["descriptions"]:
            if description["lang"] == "en":
                if "summary" in cve:
                    cve["summary"] += f" {description['value']}"
                else:
                    cve["summary"] = description["value"]

        if "metrics" in item["cve"]:
            cve["access"] = {}
            cve["impact"] = {}
            if "cvssMetricV31" in item["cve"]["metrics"]:
                cve["impact3"] = {}
                cve["exploitability3"] = {}
                cve["impact3"]["availability"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["availabilityImpact"]
                cve["impact3"]["confidentiality"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["confidentialityImpact"]
                cve["impact3"]["integrity"] = item["cve"]["metrics"]["cvssMetricV31"][
                    0
                ]["cvssData"]["integrityImpact"]
                cve["exploitability3"]["attackvector"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["attackVector"]
                cve["exploitability3"]["attackcomplexity"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["attackComplexity"]
                cve["exploitability3"]["privilegesrequired"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["privilegesRequired"]
                cve["exploitability3"]["userinteraction"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["userInteraction"]
                cve["exploitability3"]["scope"] = item["cve"]["metrics"][
                    "cvssMetricV31"
                ][0]["cvssData"]["scope"]
                cve["cvss3"] = float(
                    item["cve"]["metrics"]["cvssMetricV31"][0]["cvssData"]["baseScore"]
                )
                cve["cvss3Vector"] = item["cve"]["metrics"]["cvssMetricV31"][0][
                    "cvssData"
                ]["vectorString"]
                cve["impactScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV31"][0]["impactScore"]
                )
                cve["exploitabilityScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV31"][0]["exploitabilityScore"]
                )
                cve["cvss3Time"] = parse_datetime(
                    item["cve"]["lastModified"], ignoretz=True
                )
                cve["cvss3Type"] = item["cve"]["metrics"]["cvssMetricV31"][0]["type"]
                cve["cvss3Source"] = item["cve"]["metrics"]["cvssMetricV31"][0][
                    "source"
                ]
            elif "cvssMetricV30" in item["cve"]["metrics"]:
                cve["impact3"] = {}
                cve["exploitability3"] = {}
                cve["impact3"]["availability"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["availabilityImpact"]
                cve["impact3"]["confidentiality"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["confidentialityImpact"]
                cve["impact3"]["integrity"] = item["cve"]["metrics"]["cvssMetricV30"][
                    0
                ]["cvssData"]["integrityImpact"]
                cve["exploitability3"]["attackvector"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["attackVector"]
                cve["exploitability3"]["attackcomplexity"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["attackComplexity"]
                cve["exploitability3"]["privilegesrequired"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["privilegesRequired"]
                cve["exploitability3"]["userinteraction"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["userInteraction"]
                cve["exploitability3"]["scope"] = item["cve"]["metrics"][
                    "cvssMetricV30"
                ][0]["cvssData"]["scope"]
                cve["cvss3"] = float(
                    item["cve"]["metrics"]["cvssMetricV30"][0]["cvssData"]["baseScore"]
                )
                cve["cvss3Vector"] = item["cve"]["metrics"]["cvssMetricV30"][0][
                    "cvssData"
                ]["vectorString"]
                cve["impactScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV30"][0]["impactScore"]
                )
                cve["exploitabilityScore3"] = float(
                    item["cve"]["metrics"]["cvssMetricV30"][0]["exploitabilityScore"]
                )
                cve["cvss3Time"] = parse_datetime(
                    item["cve"]["lastModified"], ignoretz=True
                )
                cve["cvss3Type"] = item["cve"]["metrics"]["cvssMetricV30"][0]["type"]
                cve["cvss3Source"] = item["cve"]["metrics"]["cvssMetricV30"][0][
                    "source"
                ]
            else:
                cve["cvss3"] = None

            if "cvssMetricV2" in item["cve"]["metrics"]:
                cve["access"]["authentication"] = item["cve"]["metrics"][
                    "cvssMetricV2"
                ][0]["cvssData"]["authentication"]
                cve["access"]["complexity"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["accessComplexity"]
                cve["access"]["vector"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["accessVector"]
                cve["impact"]["availability"] = item["cve"]["metrics"]["cvssMetricV2"][
                    0
                ]["cvssData"]["availabilityImpact"]
                cve["impact"]["confidentiality"] = item["cve"]["metrics"][
                    "cvssMetricV2"
                ][0]["cvssData"]["confidentialityImpact"]
                cve["impact"]["integrity"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["integrityImpact"]
                cve["cvss"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["cvssData"]["baseScore"]
                )
                cve["exploitabilityScore"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["exploitabilityScore"]
                )
                cve["impactScore"] = float(
                    item["cve"]["metrics"]["cvssMetricV2"][0]["impactScore"]
                )
                cve["cvssTime"] = parse_datetime(
                    item["cve"]["lastModified"], ignoretz=True
                )  # NVD JSON lacks the CVSS time which was present in the original XML format
                cve["cvssVector"] = item["cve"]["metrics"]["cvssMetricV2"][0][
                    "cvssData"
                ]["vectorString"]
                cve["cvssType"] = item["cve"]["metrics"]["cvssMetricV2"][0]["type"]
                cve["cvssSource"] = item["cve"]["metrics"]["cvssMetricV2"][0]["source"]
            else:
                cve["cvss"] = None

        if "references" in item["cve"]:
            cve["references"] = []
            for ref in item["cve"]["references"]:
                cve["references"].append(ref["url"])

        if "configurations" in item["cve"]:
            cve["vulnerable_configuration"] = []
            cve["vulnerable_product"] = []
            cve["vendors"] = []
            cve["products"] = []
            cve["vulnerable_product_stems"] = []
            cve["vulnerable_configuration_stems"] = []
            cve["configurations"] = item["cve"]["configurations"]
            for node in item["cve"]["configurations"]:
                for cpe in node["nodes"]:
                    if "cpeMatch" in cpe:
                        for cpeuri in cpe["cpeMatch"]:
                            if "criteria" not in cpeuri:
                                continue
                            if cpeuri["vulnerable"]:
                                query = self.get_cpe_info(cpeuri)
                                if query != {}:
                                    cpe_info = sorted(
                                        self.getCPEVersionInformation(query),
                                        key=lambda x: x["padded_version"],
                                    )
                                    if cpe_info:
                                        if not isinstance(cpe_info, list):
                                            cpe_info = [cpe_info]

                                        for vulnerable_version in cpe_info:
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product",
                                                vulnerable_version["cpeName"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration",
                                                vulnerable_version["cpeName"],
                                            )
                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_configuration_stems",
                                                vulnerable_version["stem"],
                                            )

                                            cve = self.add_if_missing(
                                                cve,
                                                "vendors",
                                                vulnerable_version["vendor"],
                                            )

                                            cve = self.add_if_missing(
                                                cve,
                                                "products",
                                                vulnerable_version["product"],
                                            )

                                            cve = self.add_if_missing(
                                                cve,
                                                "vulnerable_product_stems",
                                                vulnerable_version["stem"],
                                            )
                                else:
                                    # If the cpeMatch did not have any of the version start/end modifiers,
                                    # add the CPE string as it is.
                                    cve = self.add_if_missing(
                                        cve, "vulnerable_product", cpeuri["criteria"]
                                    )
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration",
                                        cpeuri["criteria"],
                                    )
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_configuration_stems",
                                        self.stem(cpeuri["criteria"]),
                                    )
                                    vendor, product = self.get_vendor_product(
                                        cpeuri["criteria"]
                                    )
                                    cve = self.add_if_missing(cve, "vendors", vendor)
                                    cve = self.add_if_missing(cve, "products", product)
                                    cve = self.add_if_missing(
                                        cve,
                                        "vulnerable_product_stems",
                                        self.stem(cpeuri["criteria"]),
                                    )
                            else:
                                cve = self.add_if_missing(
                                    cve, "vulnerable_configuration", cpeuri["criteria"]
                                )
                                cve = self.add_if_missing(
                                    cve,
                                    "vulnerable_configuration_stems",
                                    self.stem(cpeuri["criteria"]),
                                )
        if "weaknesses" in item["cve"]:
            for problem in item["cve"]["weaknesses"]:
                for cwe in problem[
                    "description"
                ]:  # NVD JSON not clear if we can get more than one CWE per CVE (until we take the last one) -
                    # NVD-CWE-Other??? list?
                    if cwe["lang"] == "en":
                        cve["cwe"] = cwe["value"]
            if not ("cwe" in cve):
                cve["cwe"] = defaultvalue["cwe"]
        else:
            cve["cwe"] = defaultvalue["cwe"]

        cve["vulnerable_configuration_cpe_2_2"] = []

        return cve

    def process_downloads(self, sites: list = None):
        """
        Method to download and process files
        """

        self.logger.info("Starting download...")

        start_time = time.time()

        self.last_modified = datetime.datetime.now()

        self.logger.debug(
            f"do_process = {self.do_process}; is_update = {self.is_update}"
        )

        if self.do_process:
            if not self.is_update:
                try:
                    total_results = self.api_handler.get_count(
                        self.api_handler.datasource.CVE
                    )
                except ApiMaxRetryError:
                    # failed to get the count; set total_results to 0 and continue
                    total_results = 0

                self.logger.info(f"Preparing to download {total_results} CVE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(data_type="cve"):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                if not isinstance(data_list, Exception):
                                    processed_items = [
                                        self.process_item(item)
                                        for item in data_list["vulnerabilities"]
                                    ]
                                    self._db_bulk_writer(
                                        processed_items, initialization_run=True
                                    )
                                    pbar.update(len(data_list["vulnerabilities"]))
                                else:
                                    self.logger.error(
                                        f"Retrieval of api data resulted in an Exception: {data_list}..."
                                    )
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )
            else:
                last_mod_start_date = self.database[self.feed_type.lower()].find_one(
                    {}, {"lastModified": 1}, sort=[("lastModified", -1)]
                )

                if last_mod_start_date is not None:
                    if "lastModified" in last_mod_start_date:
                        last_mod_start_date = last_mod_start_date["lastModified"]
                    else:
                        raise KeyError(
                            "Missing field 'lastModified' from database query..."
                        )
                else:
                    self.logger.warning(
                        "No records found in the mongodb cves collection.."
                    )
                    return

                    # Get datetime from runtime
                last_mod_end_date = datetime.datetime.now()

                try:
                    total_results = self.api_handler.get_count(
                        self.api_handler.datasource.CVE,
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    )
                except ApiMaxRetryError:
                    # failed to get the count; set total_results to 0 and continue
                    total_results = 0

                self.logger.info(f"Preparing to download {total_results} CVE entries")

                with tqdm(
                    desc="Downloading and processing content",
                    total=total_results,
                    position=0,
                    leave=True,
                ) as pbar:
                    for entry in self.api_handler.get_all_data(
                        data_type="cve",
                        last_mod_start_date=last_mod_start_date,
                        last_mod_end_date=last_mod_end_date,
                    ):
                        # do something here with the results...
                        for data_list in tqdm(
                            entry, desc=f"Processing batch", leave=False
                        ):
                            if not isinstance(data_list, ApiDataRetrievalFailed):
                                if not isinstance(data_list, Exception):
                                    processed_items = [
                                        self.process_item(item)
                                        for item in data_list["vulnerabilities"]
                                    ]
                                    self._db_bulk_writer(processed_items)
                                    pbar.update(len(data_list["vulnerabilities"]))
                                else:
                                    self.logger.error(
                                        f"Retrieval of api data resulted in an Exception: {data_list}..."
                                    )
                            else:
                                self.logger.error(
                                    f"Retrieval of api data on url: {data_list.args[0]} failed...."
                                )

            # Set the last update time in the info collection
            self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(
            f"Duration: {datetime.timedelta(seconds=time.time() - start_time)}"
        )

    def update(self):
        self.logger.info("CVE database update started")

        self.process_downloads()

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.database_indexer.create_indexes(collection=self.feed_type.lower())
            self.is_update = False

        self.logger.info("Finished CVE database update")

        return self.last_modified

    def populate(self):
        self.logger.info("CVE database population started")

        self.logger.info(
            f"Starting CVE database population starting from year: {self.config.CVE_START_YEAR}"
        )

        self.is_update = False

        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        self.process_downloads()

        self.database_indexer.create_indexes(collection=self.feed_type.lower())

        self.logger.info("Finished CVE database population")

        return self.last_modified


class VIADownloads(JSONFileHandler):
    """
    Class processing VIA4 source files
    """

    def __init__(self):
        self.feed_type = "VIA4"
        self.prefix = "cves"
        super().__init__(
            feed_type=self.feed_type, prefix=self.prefix, logger_name=__name__
        )

        self.feed_url = self.config.SOURCES[self.feed_type.lower()]

    def file_to_queue(self, file_tuple: Tuple[str, str]):
        working_dir, filename = file_tuple

        for cve in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            x = 0
            for key, val in cve.items():
                entry_dict = {"id": key}
                entry_dict.update(val)
                self.process_item(item=entry_dict)
                x += 1

            self.logger.debug(f"Processed {x} items from file: {filename}")

        with open(filename, "rb") as input_file:
            data = json.loads(input_file.read().decode("utf-8"))

            self.setColInfo("via4", "sources", data["metadata"]["sources"])
            self.setColInfo("via4", "searchables", data["metadata"]["searchables"])

            self.logger.debug(f"Processed metadata from file: {filename}")

        try:
            self.logger.debug(f"Removing working dir: {working_dir}")
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(f"Failed to remove working dir; error produced: {err}")

    def process_item(self, item: dict):
        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    doc=item,
                )
            )

    def update(self, **kwargs):
        self.logger.info("VIA4 database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        if self.source_changed(self.feed_url):
            self.process_downloads([self.feed_url])
            self.logger.info("Finished VIA4 database update")
        else:
            self.logger.info("Skipped VIA4 database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


class CAPECDownloads(XMLFileHandler):
    """
    Class processing CAPEC source files
    """

    def __init__(self):
        self.feed_type = "CAPEC"
        super().__init__(feed_type=self.feed_type, logger_name=__name__)

        self.feed_url = self.config.SOURCES[self.feed_type.lower()]

        # make parser
        self.parser = make_parser()
        self.ch = CapecHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple: Tuple[str, str]):
        working_dir, filename = file_tuple

        self.parser.parse(filename)
        x = 0
        for attack in self.ch.capec:
            self.process_item(attack)
            x += 1

        self.logger.debug(f"Processed {x} entries from file: {filename}")

        try:
            self.logger.debug(f"Removing working dir: {working_dir}")
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(f"Failed to remove working dir; error produced: {err}")

    def update(self, **kwargs):
        self.logger.info("CAPEC database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        if self.source_changed(self.feed_url):
            self.process_downloads([self.feed_url])
            self.logger.info("Finished CAPEC database update")
        else:
            self.logger.info("Skipped CAPEC database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


class CWEDownloads(XMLFileHandler):
    """
    Class processing CWE source files
    """

    def __init__(self):
        self.feed_type = "CWE"
        super().__init__(feed_type=self.feed_type, logger_name=__name__)

        self.feed_url = self.config.SOURCES[self.feed_type.lower()]

        # make parser
        self.parser = make_parser()
        self.ch = CWEHandler()
        self.parser.setContentHandler(self.ch)

    def file_to_queue(self, file_tuple: Tuple[str, str]):
        working_dir, filename = file_tuple

        for f in glob.glob(f"{working_dir}/*.xml"):
            filename = f

        self.parser.parse(f"file://{filename}")
        x = 0
        for cwe in self.ch.cwe:
            try:
                cwe["related_weaknesses"] = list(set(cwe["related_weaknesses"]))
                cwe["description"] = cwe["Description"]
                cwe.pop("Description")
            except KeyError:
                pass
            self.process_item(cwe)
            x += 1

        self.logger.debug(f"Processed {x} entries from file: {filename}")

        try:
            self.logger.debug(f"Removing working dir: {working_dir}")
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(f"Failed to remove working dir; error produced: {err}")

    def update(self, **kwargs):
        self.logger.info("CWE database update started")

        # if collection is non-existent; assume it's not an update
        if self.feed_type.lower() not in self.getTableNames():
            self.is_update = False

        if self.source_changed(self.feed_url):
            self.process_downloads([self.feed_url])
            self.logger.info("Finished CWE database update")
        else:
            self.logger.info("Skipped CWE database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.is_update = False
        self.queue.clear()

        self.delColInfo(self.feed_type.lower())

        self.dropCollection(self.feed_type.lower())

        return self.update()


class EPSSDownloads(CSVFileHandler):
    def __init__(self):
        self.feed_type = "EPSS"
        self.delimiter = ","
        super().__init__(
            feed_type=self.feed_type, logger_name=__name__, delimiter=self.delimiter
        )

        self.feed_url = self.config.SOURCES[self.feed_type.lower()]

        self.is_update = True

    def process_epss_item(self, item=None):
        if item is None:
            return None

        epss = {
            "id": item[0],
            "epss": item[1],
            "epssMetric": {"percentile": item[2], "lastModified": self.last_modified},
        }

        return epss

    def process_item(self, item):
        epss = self.process_epss_item(item)

        if epss is not None:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne, doc=epss, upsert=False
                )
            )

    def update(self, **kwargs):
        self.logger.info("EPSS database update started")

        self.queue.clear()

        self.process_downloads([self.feed_url])

        self.logger.info("Finished EPSS database update")

        return self.last_modified

    def populate(self, **kwargs):
        self.logger.info("EPSS Database population started")

        self.queue.clear()

        self.process_downloads([self.feed_url])

        self.logger.info("Finished EPSS database population")

        return self.last_modified



import re
from functools import reduce
from typing import Union, Iterable

from CveXplore.common.data_source_connection import DatasourceConnection
from CveXplore.common.db_mapping import database_mapping
from CveXplore.objects.cvexplore_object import CveXploreObject


class GenericDatabaseFactory(DatasourceConnection):
    """
    The GenericDatabaseFactory handles the creation of general, collection based, functions which provide an instance
    of CveXplore functions that apply to the given collection.
    """

    def __init__(self, collection: str):
        """
        Create a new GenericDatabaseFactory and create field specific functions based on the __default_fields and
        the __fields_mapping.
        """
        super().__init__(collection=collection)

        self.__database_mapping = database_mapping

        self.__default_fields = ["id"]

        self.__fields_mapping = {
            "capec": [
                "name",
                "summary",
                "prerequisites",
                "solutions",
                "loa",
                "typical_severity",
            ],
            "cpe": ["title", "cpeName", "vendor", "product", "stem"],
            "cwe": ["name", "status", "description"],
            "cves": [
                "cvss",
                "cvss3",
                "summary",
                "vendors",
                "products",
                "lastModified",
                "modified",
                "published",
                "status",
                "assigner",
                "cwe",
                "epss",
            ],
        }

        total_fields_list = (
            self.__default_fields + self.__fields_mapping[self.collection]
        )
        for field in total_fields_list:
            setattr(
                self,
                field,
                GenericDatabaseFieldsFunctions(field=field, collection=self.collection),
            )

    def get_by_id(self, doc_id: str):
        """
        Method to fetch a specific collection entry via it's id number
        """

        if not isinstance(doc_id, str):
            try:
                doc_id = str(doc_id)
            except ValueError:
                return "Provided value is not a string nor can it be cast to one"

        return self.datasource_collection_connection.find_one({"id": doc_id})

    def mget_by_id(self, *doc_ids: str) -> Union[Iterable[CveXploreObject], Iterable]:
        """
        Method to fetch a specific collection entry via it's id number
        """
        ret_data = []
        for doc_id in doc_ids:
            data = self.get_by_id(doc_id=doc_id)
            if data is not None:
                ret_data.append(data)

        if len(ret_data) >= 1:
            return sorted(ret_data, key=lambda x: x.id)
        else:
            return ret_data

    def _field_list(self, doc_id: str) -> list:
        """
        Method to fetch all field names from a specific collection
        """
        return sorted(
            list(
                reduce(
                    lambda all_keys, rec_keys: all_keys | set(rec_keys),
                    map(
                        lambda d: d.to_dict(),
                        [
                            self.datasource_collection_connection.find_one(
                                {"id": doc_id}
                            )
                        ],
                    ),
                    set(),
                )
            )
        )

    def field_list(self, *doc_ids: str) -> list:
        """
        Method to fetch all field names from a specific collection
        """
        ret_data = []
        for doc_id in sorted(doc_ids):
            data = self._field_list(doc_id=doc_id)
            ret_data.append({doc_id: data, "field_count": len(data)})

        return ret_data

    def mapped_fields(self, collection: str) -> list:
        ret_list = []
        try:
            ret_list.extend(self.__fields_mapping[collection])
        except KeyError:
            pass

        ret_list.extend(self.__default_fields)
        return sorted(ret_list)

    def __repr__(self):
        """String representation of object"""
        return f"<< {self.__class__.__name__}:{self.collection} >>"


class GenericDatabaseFieldsFunctions(DatasourceConnection):
    """
    The GenericDatabaseFieldsFunctions handles the creation of general, field based, functions
    """

    def __init__(self, field: str, collection: str):
        """
        Create a new GenericDatabaseFieldsFunctions and create field specific functions.
        """
        super().__init__(collection=collection)

        self.__field = field

    def search(self, value: str):
        """
        Method for searching for a given value. The value shall be converted to a regex.
        """

        regex = re.compile(value, re.IGNORECASE)

        query = {self.__field: {"$regex": regex}}

        return self.datasource_collection_connection.find(query)

    def find(self, value: str | dict = None):
        """
        Method to find a given value.
        """

        if value is not None:
            query = {self.__field: value}
        else:
            query = None

        return self.datasource_collection_connection.find(query)

    def __repr__(self):
        """String representation of object"""
        return f"<< GenericDatabaseFieldsFunctions:{self.collection} >>"



from CveXplore.api.connection.api_db import ApiDatabaseSource
from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase
from CveXplore.database.connection.dummy.dummy import DummyConnection
from CveXplore.database.connection.mongodb.mongo_db import MongoDBConnection
from CveXplore.database.connection.sqlbase.sql_base import SQLBaseConnection


class DatabaseConnection(object):
    def __init__(self, database_type: str, database_init_parameters: dict):
        self.database_type = database_type
        self.database_init_parameters = database_init_parameters

        self._database_connections = {
            "mongodb": MongoDBConnection,
            "api": ApiDatabaseSource,
            "mysql": SQLBaseConnection,
            "dummy": DummyConnection,
        }

        self._database_connection = self._database_connections[self.database_type](
            **self.database_init_parameters
        )

    @property
    def database_connection(self) -> DatabaseConnectionBase:
        return self._database_connection



from CveXplore.common.data_source_connection import DatasourceConnection


class Capec(DatasourceConnection):
    """
    Capec database object
    """

    def __init__(self, **kwargs):
        super().__init__("capec")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def iter_related_weaknessess(self):
        """
        Generator function for iterating the related weaknesses from the current Capec object

        :return: Cwe object
        :rtype: Cwe
        """
        if hasattr(self, "related_weakness"):
            if len(self.related_weakness) != 0:
                for each in self.related_weakness:
                    cwe_doc = self.datasource_connection.store_cwe.find_one(
                        {"id": each}
                    )

                    yield cwe_doc

    def iter_related_capecs(self):
        """
        Generator function for iterating the related capecs from the current Capec object

        :return: Capec objects
        :rtype: Capec
        """
        if hasattr(self, "related_capecs"):
            if len(self.related_capecs) != 0:
                for each in self.related_capecs:
                    capec_doc = self.datasource_connection.store_capec.find_one(
                        {"id": each}
                    )

                    yield capec_doc

    def __repr__(self):
        """String representation of object"""
        return f"<< Capec:{self.id} >>"



import csv
import shutil
from abc import abstractmethod
from typing import Tuple

from CveXplore.core.database_actions.db_action import DatabaseAction
from CveXplore.core.database_maintenance.download_handler import DownloadHandler
from CveXplore.core.database_maintenance.ijson_handler import IJSONHandler


class JSONFileHandler(DownloadHandler):
    """
    This class handles all JSON related download processing and functions as a base class for specific JSON sources
    processing and downloading
    """

    def __init__(self, feed_type: str, prefix: str, logger_name: str):
        super().__init__(feed_type=feed_type, prefix=prefix, logger_name=logger_name)

        self.is_update = True

        self.prefix = prefix

        self.ijson_handler = IJSONHandler()

    def file_to_queue(self, file_tuple: Tuple[str, str]):
        """
        Method responsible for transferring file contents to the worker queue for further processing and inserting them
        into the database
        """

        working_dir, filename = file_tuple

        # adjust the interval counter for debug logging when updating
        if self.is_update:
            interval = 500
        else:
            interval = 5000

        x = 0
        self.logger.debug(f"Starting processing of file: {filename}")
        for cpe in self.ijson_handler.fetch(filename=filename, prefix=self.prefix):
            self.process_item(item=cpe)
            x += 1
            if x % interval == 0:
                self.logger.debug(f"Processed {x} entries from file: {filename}")

        try:
            self.logger.debug(f"Removing working dir: {working_dir}")
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(f"Failed to remove working dir; error produced: {err}")

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError


class XMLFileHandler(DownloadHandler):
    """
    This class handles all XML related download processing and functions as a base class for specific XML sources
    processing and downloading
    """

    def __init__(self, feed_type: str, logger_name: str):
        super().__init__(feed_type=feed_type, logger_name=logger_name)
        self.is_update = True

    def process_item(self, item: dict):
        """
        Method responsible for putting items into the worker queue as database actions
        """

        if self.is_update:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    doc=item,
                )
            )
        else:
            self.queue.put(
                DatabaseAction(
                    action=DatabaseAction.actions.InsertOne,
                    doc=item,
                )
            )

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError


class CSVFileHandler(DownloadHandler):
    def __init__(self, feed_type, logger_name: str, delimiter=","):
        super().__init__(feed_type=feed_type, logger_name=logger_name)

        self.is_update = True
        self.delimiter = delimiter

    def file_to_queue(self, file_tuple):
        working_dir, filename = file_tuple

        # adjust the interval counter for debug logging when updating
        if self.is_update:
            interval = 500
        else:
            interval = 5000

        x = 0
        self.logger.debug(f"Starting processing of file: {filename}")

        f = open(filename, "r")

        reader = csv.reader(f, delimiter=self.delimiter)

        next(reader)
        next(reader)

        for row in reader:
            self.process_item(item=row)
            x += 1
            if x % interval == 0:
                self.logger.debug(f"Processed {x} entries from file: {filename}")
        f.close()

        try:
            self.logger.debug(f"Removing working dir: {working_dir}")
            shutil.rmtree(working_dir)
        except Exception as err:
            self.logger.error(f"Failed to remove working dir; error produced: {err}")

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



from abc import abstractmethod

from CveXplore.common.cpe_converters import split_cpe_name
from CveXplore.core.database_actions.db_action import DatabaseAction
from CveXplore.core.database_maintenance.download_handler import DownloadHandler
from CveXplore.core.nvd_nist.nvd_nist_api import NvdNistApi


class NVDApiHandler(DownloadHandler):
    """
    This class handles all JSON related download processing and functions as a base class for specific JSON sources
    processing and downloading
    """

    def __init__(self, feed_type: str, logger_name: str):
        super().__init__(feed_type=feed_type, logger_name=logger_name)

        self.is_update = True

        self.api_handler = NvdNistApi(proxies=self.config.HTTP_PROXY_DICT)

    def process_item(self, item: dict):
        item = self.process_the_item(item)

        if item is not None:
            if self.is_update:
                return DatabaseAction(
                    action=DatabaseAction.actions.UpdateOne,
                    doc=item,
                ).entry
            else:
                # return DatabaseAction(
                #     action=DatabaseAction.actions.InsertOne,
                #     doc=item,
                # ).entry
                return item

    @staticmethod
    def split_cpe_name(cpename: str) -> list[str]:
        return split_cpe_name(cpename)

    def stem(self, cpe_uri: str):
        cpe_stem = self.split_cpe_name(cpe_uri)
        return ":".join(cpe_stem[:5])

    @staticmethod
    def padded_version(version: str):
        if version == "-" or version == "":
            return version
        else:
            # normalizing edge cases:
            version = version.replace("\\(", ".").replace("\\)", ".").rstrip(".")

            ret_list = []

            splitted_version = version.split(".")
            # perform check if last part of version can be cast to an int
            try:
                int(splitted_version[-1])
                # can be cast to an int, proceed 'normally'
                for v in splitted_version:
                    try:
                        ret_list.append(f"{int(v):05d}")
                    except ValueError:
                        ret_list.append(v.rjust(5, "0"))
            except ValueError:
                # last part of the version cannot be cast to an int, so this means it's either a string or a
                # string combined with an integer; handle accordingly

                # first handle all version identifiers leading upto the last part
                if len(splitted_version) > 1:
                    for i in range(len(splitted_version) - 1):
                        try:
                            ret_list.append(f"{int(splitted_version[i]):05d}")
                        except ValueError:
                            ret_list.append(splitted_version[i].rjust(5, "0"))

                # handle the last part
                # check if the last entry is smaller than 5 characters, if so just use that...
                if len(splitted_version[-1]) > 5:
                    try:
                        ret_list.append(f"{int(splitted_version[-1]):05d}")
                    except ValueError:
                        ret_list.append(splitted_version[-1].rjust(5, "0"))
                # check is last entry consists only of alphanumeric characters
                elif splitted_version[-1].isalpha():
                    ret_list.append(splitted_version[-1].rjust(5, "0"))
                else:
                    loop_i = 0
                    loop_count = len(splitted_version[-1])

                    # int/str combined value; handle accordingly
                    while loop_i < loop_count:
                        current_i = loop_i
                        # probably digit; so check;
                        if splitted_version[-1][loop_i].isdigit():
                            try:
                                ret_list.append(
                                    f"{int(splitted_version[-1][loop_i]):05d}"
                                )
                            except ValueError:
                                ret_list.append(
                                    splitted_version[-1][loop_i].rjust(5, "0")
                                )
                            finally:
                                # perform check if anything that follows consists only of string characters
                                if splitted_version[-1][loop_i + 1 :].isalpha():
                                    ret_list.append(
                                        splitted_version[-1][loop_i + 1 :].rjust(5, "0")
                                    )
                                    # no point proceeding; just break
                                    break
                                loop_i += 1
                        else:
                            # ok so probably last part of version identifier is a string; add that with a loop
                            version_string = ""
                            try:
                                while splitted_version[-1][loop_i].isalpha():
                                    version_string += splitted_version[-1][loop_i]
                                    loop_i += 1
                            except IndexError:
                                # finished splitted_version variable; just pass
                                loop_i += 1
                                pass

                            ret_list.append(version_string.rjust(5, "0"))

                        if loop_i == current_i:
                            loop_i += 1

            return ".".join(ret_list)

    @abstractmethod
    def process_the_item(self, *args):
        raise NotImplementedError

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



import json
import logging
import os
import time
from datetime import timedelta

from CveXplore.core.database_indexer.db_indexer import DatabaseIndexer
from CveXplore.core.database_maintenance.sources_process import (
    CPEDownloads,
    CVEDownloads,
    CWEDownloads,
    CAPECDownloads,
    VIADownloads,
    EPSSDownloads,
)
from CveXplore.core.database_maintenance.update_base_class import UpdateBaseClass
from CveXplore.core.database_migration.database_migrator import DatabaseMigrator
from CveXplore.core.database_version.db_version_checker import DatabaseVersionChecker
from CveXplore.core.logging.logger_class import AppLogger
from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase
from CveXplore.errors import UpdateSourceNotFound

logging.setLoggerClass(AppLogger)


class MainUpdater(UpdateBaseClass):
    """
    The MainUpdater class is the main class for performing database database_maintenance tasks
    """

    def __init__(self, datasource: DatabaseConnectionBase, datasource_type: str):
        """
        Init a new MainUpdater class
        """
        super().__init__(logger_name=__name__)

        self.datasource = datasource
        self.datasource_type = datasource_type

        self.sources = [
            {"name": "cpe", "updater": CPEDownloads},
            {"name": "cve", "updater": CVEDownloads},
            {"name": "cwe", "updater": CWEDownloads},
            {"name": "capec", "updater": CAPECDownloads},
            {"name": "via4", "updater": VIADownloads},
            {"name": "epss", "updater": EPSSDownloads},
        ]

        self.database_indexer = DatabaseIndexer(datasource=datasource)
        self.schema_checker = DatabaseVersionChecker(datasource=datasource)
        self.database_migrator = DatabaseMigrator()

        self.do_initialize = False

    def __repr__(self):
        return f"<<MainUpdater>>"

    def validate_schema(self):
        return self.schema_checker.validate_schema()

    def reset_download_sources_to_default(self):
        sources = self.config.DEFAULT_SOURCES

        with open(os.path.join(self.config.USER_HOME_DIR, ".sources.ini"), "w") as f:
            f.write(json.dumps(sources))

        return True

    def update(self, update_source: str | list = None):
        """
        Method used for updating the database
        """
        self.logger.info(f"Starting Database update....")
        start_time = time.time()

        if not self.do_initialize:
            if self.datasource_type != "api" and self.datasource_type != "mongodb":
                self.logger.info(
                    f"Upgrading database schema to latest head, if needed..."
                )
                self.database_migrator.db_upgrade()

        if update_source is not None:
            if not isinstance(update_source, str | list):
                raise ValueError("Wrong 'update_source' parameter type received!")

        try:
            if update_source is None:
                for source in self.sources:
                    up = source["updater"]()
                    up.update()

            elif isinstance(update_source, list):
                for source in update_source:
                    try:
                        update_this_source = [
                            x for x in self.sources if x["name"] == source
                        ][0]
                        up = update_this_source["updater"]()
                        up.update()
                    except IndexError:
                        raise UpdateSourceNotFound(
                            f"Provided source: {source} could not be found...."
                        )
            else:
                # single string then....
                try:
                    update_this_source = [
                        x for x in self.sources if x["name"] == update_source
                    ][0]
                    up = update_this_source["updater"]()
                    up.update()
                except IndexError:
                    raise UpdateSourceNotFound(
                        f"Provided source: {update_source} could not be found...."
                    )
        except UpdateSourceNotFound:
            raise

        self.database_indexer.create_indexes()

        self.schema_checker.update()

        self.datasource.set_handlers_for_collections()

        self.logger.info(f"Database update complete!")
        self.logger.info(
            f"Update Total duration: {timedelta(seconds=time.time() - start_time)}"
        )

    def populate(self, populate_source: str | list = None):
        """
        Method used for updating the database
        """
        self.logger.info(f"Starting Database population....")
        start_time = time.time()

        if self.datasource_type != "api" and self.datasource_type != "mongodb":
            self.logger.info(f"Upgrading database schema to latest head, if needed...")
            self.database_migrator.db_upgrade()

        if populate_source is not None:
            if not isinstance(populate_source, str | list):
                raise ValueError("Wrong 'populate_source' parameter type received!")

        try:
            if populate_source is None:
                for source in self.sources:
                    up = source["updater"]()
                    up.populate()

            elif isinstance(populate_source, list):
                for source in populate_source:
                    try:
                        update_this_source = [
                            x for x in self.sources if x["name"] == source
                        ][0]
                        up = update_this_source["updater"]()
                        up.populate()
                    except IndexError:
                        raise UpdateSourceNotFound(
                            f"Provided source: {source} could not be found...."
                        )
            else:
                # single string then....
                try:
                    update_this_source = [
                        x for x in self.sources if x["name"] == populate_source
                    ][0]
                    up = update_this_source["updater"]()
                    up.populate()
                except IndexError:
                    raise UpdateSourceNotFound(
                        f"Provided source: {populate_source} could not be found...."
                    )
        except UpdateSourceNotFound:
            raise

        self.database_indexer.create_indexes()

        self.schema_checker.update()

        self.datasource.set_handlers_for_collections()

        self.logger.info(f"Database population complete!")
        self.logger.info(
            f"Populate total duration: {timedelta(seconds=time.time() - start_time)}"
        )

    def initialize(self):
        """
        Method to initialize a new (fresh) instance of a cvedb database
        """

        self.logger.info(f"Starting Database initialization....")
        start_time = time.time()

        self.do_initialize = True

        if self.datasource_type != "api" and self.datasource_type != "mongodb":
            self.logger.info(f"Upgrading database schema to latest head, if needed...")
            self.database_migrator.db_upgrade()

        cpe_pop = CPEDownloads()
        cpe_pop.populate()

        self.logger.info(
            f"Sleeping for 30 seconds between CPE and CVE database population.."
        )
        time.sleep(30)

        cve_pop = CVEDownloads()
        cve_pop.populate()

        self.update()

        self.logger.info(f"Database initialization complete!")
        self.logger.info(
            f"Initialization total duration: {timedelta(seconds=time.time() - start_time)}"
        )



import collections

from pymongo import InsertOne, UpdateOne


class DatabaseAction(object):
    """
    Database Action class is used to queue update and insert actions.

    Group:
        database
    """

    actions = collections.namedtuple("Actions", "InsertOne UpdateOne")(0, 1)

    def __init__(self, action: actions, doc: dict, upsert: bool = True):
        """
        Create a DatabaseAction object.

        Args:
            action: Action to use
            doc: Entry to perform action with
        """
        self.action = action
        self.doc = doc
        self.upsert = upsert

    @property
    def entry(self):
        """
        Show the entry property for this class

        Returns:
            The entry property for this class.
        Group:
            properties
        """
        if self.action == self.actions.InsertOne:
            return InsertOne(self.doc)
        elif self.action == self.actions.UpdateOne:
            return UpdateOne(
                {"id": self.doc["id"]}, {"$set": self.doc}, upsert=self.upsert
            )



"""First 2 models

Revision ID: 53df0e286532
Revises: 
Create Date: 2023-12-20 16:18:04.244419

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "53df0e286532"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "info",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("db", sa.String(length=25), nullable=True),
        sa.Column("lastModified", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "schema",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("rebuild_needed", sa.Boolean(), nullable=True),
        sa.Column("version", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("schema")
    op.drop_table("info")
    # ### end Alembic commands ###



import collections

task_status_codes = collections.namedtuple("task_status_codes", ["OK", "NOK"])(0, 1)

task_status_list = [
    x.lower()
    for x in task_status_codes.__dir__()
    if not x.startswith("_") and x not in ["index", "count"]
]

task_status_rev_types = {
    getattr(task_status_codes, x.upper()): x for x in task_status_list
}



import calendar
from datetime import datetime

import colors
import pymongo


def sanitize(x: list | pymongo.cursor.Cursor):
    if isinstance(x, pymongo.cursor.Cursor):
        x = list(x)
    if type(x) == list:
        for y in x:
            sanitize(y)
    if x and "_id" in x:
        x.pop("_id")
    return x


def timestampTOdatetime(timestamp):
    """
    Method that will take the provided timestamp and converts it into a date time object

    :param timestamp: unix timestamp
    :type timestamp: int
    :return: date time object
    :rtype: datetime
    """
    value = datetime.utcfromtimestamp(timestamp)

    return value


def datetimeTOtimestamp(date_time_object):
    return calendar.timegm(date_time_object.utctimetuple())


def datetimeToTimestring(datetime_object):
    return timestampTOdatetimestring(datetimeTOtimestamp(datetime_object))


def timestampTOdatetimestring(timestamp, vis=False):
    """
    Method that will take the provided timestamp and converts it into a RFC3339 date time string

    :param timestamp: unix timestamp
    :type timestamp: int
    :return: date time object
    :rtype: datetime.datetime (format: '%d-%m-%YT%H:%M:%SZ')
    """
    value = datetime.utcfromtimestamp(timestamp)

    if not vis:
        return value.strftime("%Y-%m-%dT%H:%M:%SZ")
    else:
        return value.strftime("%Y-%m-%d %H:%M:%S")


def set_ansi_color_red(msg: str) -> str:
    return colors.color("{}".format(msg), fg="red")


def set_ansi_color_green(msg: str) -> str:
    return colors.color("{}".format(msg), fg="green")


def set_ansi_color_magenta(msg: str) -> str:
    return colors.color("{}".format(msg), fg="magenta")


def set_ansi_color_yellow(msg: str) -> str:
    return colors.color("{}".format(msg), fg="yellow")



from sqlalchemy import (
    Column,
    Integer,
    String,
    DateTime,
    Boolean,
    Float,
    JSON,
    Text,
    BigInteger,
)

from sqlalchemy.orm import declarative_base

CveXploreBase = declarative_base()


class Info(CveXploreBase):
    """
    Information about the time the last update was made to a db table.
    """

    __tablename__ = "info"
    id = Column(Integer, primary_key=True, doc="ID of the record")
    db = Column(String(25), doc="Database name")
    lastModified = Column(DateTime, doc="Last modified time")

    def __repr__(self):
        return f"<< Info: {self.db} >>"


class Schema(CveXploreBase):
    """
    Holds information about the current database schema.
    """

    __tablename__ = "schema"
    id = Column(Integer, primary_key=True, doc="ID of the record")
    rebuild_needed = Column(
        Boolean, default=False, doc="Flag to indicate if a rebuild is needed"
    )
    version = Column(Float, doc="Version of the database schema")

    def __repr__(self):
        return f"<< Schema: {self.id} >>"


class Capec(CveXploreBase):
    """
    CAPEC database table.
    """

    __tablename__ = "capec"
    id = Column(
        Integer, primary_key=True, unique=True, index=True, doc="ID of the CAPEC"
    )
    loa = Column(String(25), index=True, doc="Likelihood Of Attack")
    name = Column(String(250), index=True, doc="Name of the CAPEC")
    prerequisites = Column(Text, doc="Prerequisites of the CAPEC")
    solutions = Column(Text, doc="Solutions related to the CAPEC")
    summary = Column(Text, doc="Summary of the CAPEC")
    typical_severity = Column(
        String(25), index=True, doc="Typical severity of the CAPEC"
    )
    execution_flow = Column(JSON, default={}, doc="Execution flow of the CAPEC")
    related_capecs = Column(JSON, default=[], doc="Related CAPECs")
    related_weakness = Column(JSON, default=[], doc="Related weakness of the CAPEC")
    taxonomy = Column(JSON, default=[], doc="Taxonomy of the CAPEC")

    def __repr__(self):
        return f"<< Capec: {self.id} >>"


class Cpe(CveXploreBase):
    """
    CPE database table.
    """

    __tablename__ = "cpe"
    _id = Column(
        BigInteger, primary_key=True, unique=True, index=True, doc="ID of the record"
    )
    id = Column(String(50), unique=True, index=True, doc="ID of the CPE")
    cpeName = Column(String(50), index=True, doc="Name of the CPE")
    cpeNameId = Column(String(50), doc="Unique ID number of the CPE")
    created = Column(DateTime, doc="Creation time of the CPE")
    deprecated = Column(
        Boolean,
        default=False,
        index=True,
        doc="Flag to indicate if a CPE is deprecated",
    )
    deprecatedBy = Column(String(50), doc="Organization that deprecated the CPE")
    lastModified = Column(DateTime, index=True, doc="Last modified time of the CPE")
    padded_version = Column(
        String(50), index=True, doc="Left zero padded version of the CPE"
    )
    product = Column(String(50), index=True, doc="Product of the CPE")
    stem = Column(String(50), index=True, doc="Stem of the CPE")
    title = Column(String(150), index=True, doc="Title of the CPE")
    vendor = Column(String(50), index=True, doc="Vendor of the CPE")
    version = Column(String(50), doc="Version of the CPE")

    def __repr__(self):
        return f"<< Cpe: {self.id} >>"


class Cves(CveXploreBase):
    """
    CVE database table.
    """

    __tablename__ = "cves"
    _id = Column(
        BigInteger, primary_key=True, unique=True, index=True, doc="ID of the record"
    )
    id = Column(String(50), unique=True, index=True, doc="ID of the CVE")
    access = Column(JSON, default={}, doc="Access information of the CVE")
    assigner = Column(
        String(50), index=True, doc="Organization name that assigned the CVE"
    )
    cvss = Column(Float, index=True, doc="CVSS of the CVE")
    cvss3 = Column(Float, index=True, doc="CVSS3 of the CVE")
    cvssSource = Column(String(50), doc="Source of the CVSS of the CVE")
    cvssTime = Column(DateTime, doc="Time of the CVSS of the CVE")
    cvssVector = Column(String(100), doc="Vector of the CVSS of the CVE")
    configurations = Column(JSON, doc="Vulnerable configurations of the CVE")
    cwe = Column(String(50), index=True, doc="Related CWEs to the CVE")
    epss = Column(Float, index=True, doc="Epss of the CVE")
    epssMetric = Column(JSON, doc="Epss metric of the CVE")
    exploitabilityScore = Column(Float, doc="Exploitability Score of the CVE")
    impact = Column(JSON, doc="Impact of the CVE")
    impactScore = Column(Float, doc="Impact score of the CVE")
    lastModified = Column(DateTime, index=True, doc="Last modified time of the CVE")
    modified = Column(DateTime, index=True, doc="Modified time of the CVE")
    products = Column(JSON, default=[], doc="Related products to the CVE")
    published = Column(DateTime, index=True, doc="Published time of the CVE")
    references = Column(JSON, doc="References associated to the CVE")
    status = Column(String(25), index=True, doc="Status of the CVE")
    summary = Column(Text, doc="Summary of the CVE")
    vendors = Column(JSON, default=[], doc="Related vendors of the CVE")
    vulnerable_configuration = Column(
        JSON, default=[], doc="Vulnerable configurations to the CVE"
    )
    vulnerable_configuration_cpe_2_2 = Column(
        JSON, default=[], doc="Vulnerable configurations to the CVE"
    )
    vulnerable_configuration_stems = Column(
        JSON, default=[], doc="Vulnerable configuration stems to the CVE"
    )
    vulnerable_product = Column(JSON, default=[], doc="Vulnerable products to the CVE")
    vulnerable_product_stems = Column(
        JSON, default=[], doc="Vulnerable product stems to the CVE"
    )

    def __repr__(self):
        return f"<< Cves: {self.id} >>"


class Cwe(CveXploreBase):
    """
    CWE database table.
    """

    __tablename__ = "cwe"
    id = Column(
        Integer, primary_key=True, unique=True, index=True, doc="ID of the record"
    )
    description = Column(Text, doc="Description of the CWE")
    name = Column(String(250), index=True, doc="Name of the CWE")
    status = Column(String(25), index=True, doc="Status of the CWE")
    weaknessabs = Column(String(25), doc="Weaknessabs of the CWE")
    related_weaknesses = Column(JSON, default=[], doc="Related weaknesses to the CWE")

    def __repr__(self):
        return f"<< Cwe: {self.id} >>"


class Via4(CveXploreBase):
    """
    VIA4 database table.
    """

    __tablename__ = "via4"
    _id = Column(
        Integer, primary_key=True, unique=True, index=True, doc="ID of the record"
    )
    id = Column(String(50), index=True, doc="ID of the Via4")
    db = Column(String(25), doc="Database name")
    searchables = Column(JSON, default=[], doc="VIA4 searchables")
    sources = Column(JSON, default=[], doc="VIA4 sources")
    msbulletin = Column(JSON, default=[], doc="VIA4 msbulletin")
    redhat = Column(JSON, default={}, doc="VIA4 redhat")
    refmap = Column(JSON, default={}, doc="VIA4 refmap")

    def __repr__(self):
        return f"<< Via4: {self.db} >>"


class Cpeother(CveXploreBase):
    __tablename__ = "cpeother"
    id = Column(Integer, primary_key=True, unique=True, index=True)

    def __repr__(self):
        return f"<< Cpeother: {self.id} >>"


class MgmtBlacklist(CveXploreBase):
    __tablename__ = "mgmt_blacklist"
    id = Column(Integer, primary_key=True, unique=True, index=True)

    def __repr__(self):
        return f"<< MgmtBlacklist: {self.id} >>"


class MgmtWhitelist(CveXploreBase):
    __tablename__ = "mgmt_whitelist"
    id = Column(Integer, primary_key=True, unique=True, index=True)

    def __repr__(self):
        return f"<< MgmtWhitelist: {self.id} >>"



from collections import namedtuple

from pymongo import TEXT, ASCENDING

from CveXplore.core.database_maintenance.update_base_class import UpdateBaseClass
from CveXplore.core.general.utils import sanitize
from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase

MongoUniqueIndex = namedtuple("MongoUniqueIndex", "index name unique")
MongoAddIndex = namedtuple("MongoAddIndex", "index name")


class DatabaseIndexer(UpdateBaseClass):
    """
    Class processing the Mongodb indexes
    """

    def __init__(self, datasource: DatabaseConnectionBase):
        super().__init__(logger_name=__name__)

        database = datasource
        self.database = database.dbclient

        self.indexes = {
            "cpe": [
                MongoUniqueIndex(index=[("id", ASCENDING)], name="id", unique=True),
                MongoAddIndex(index=[("vendor", ASCENDING)], name="vendor"),
                MongoAddIndex(index=[("product", ASCENDING)], name="product"),
                MongoAddIndex(index=[("deprecated", ASCENDING)], name="deprecated"),
                MongoAddIndex(index=[("cpeName", ASCENDING)], name="cpeName"),
                MongoAddIndex(index=[("title", ASCENDING)], name="title"),
                MongoAddIndex(index=[("stem", ASCENDING)], name="stem"),
                MongoAddIndex(
                    index=[("padded_version", ASCENDING)], name="padded_version"
                ),
                MongoAddIndex(index=[("lastModified", ASCENDING)], name="lastModified"),
            ],
            "cpeother": [
                MongoUniqueIndex(index=[("id", ASCENDING)], name="id", unique=True)
            ],
            "cves": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id"),
                MongoAddIndex(
                    index=[("vulnerable_configuration", ASCENDING)],
                    name="vulnerable_configuration",
                ),
                MongoAddIndex(
                    index=[("vulnerable_product", ASCENDING)], name="vulnerable_product"
                ),
                MongoAddIndex(index=[("modified", ASCENDING)], name="modified"),
                MongoAddIndex(index=[("published", ASCENDING)], name="published"),
                MongoAddIndex(index=[("lastModified", ASCENDING)], name="lastModified"),
                MongoAddIndex(index=[("cvss", ASCENDING)], name="cvss"),
                MongoAddIndex(index=[("cvss3", ASCENDING)], name="cvss3"),
                MongoAddIndex(index=[("summary", TEXT)], name="summary"),
                MongoAddIndex(index=[("vendors", ASCENDING)], name="vendors"),
                MongoAddIndex(index=[("products", ASCENDING)], name="products"),
                MongoAddIndex(index=[("assigner", ASCENDING)], name="assigner"),
                MongoAddIndex(index=[("cwe", ASCENDING)], name="cwe"),
                MongoAddIndex(index=[("status", ASCENDING)], name="status"),
                MongoAddIndex(
                    index=[("vulnerable_product_stems", ASCENDING)],
                    name="vulnerable_product_stems",
                ),
                MongoAddIndex(
                    index=[("vulnerable_configuration_stems", ASCENDING)],
                    name="vulnerable_configuration_stems",
                ),
                MongoAddIndex(index=[("epss", ASCENDING)], name="epss"),
            ],
            "via4": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "mgmt_whitelist": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "mgmt_blacklist": [MongoAddIndex(index=[("id", ASCENDING)], name="id")],
            "capec": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id"),
                MongoAddIndex(index=[("loa", ASCENDING)], name="loa"),
                MongoAddIndex(
                    index=[("typical_severity", ASCENDING)], name="typical_severity"
                ),
                MongoAddIndex(index=[("name", ASCENDING)], name="name"),
                MongoAddIndex(
                    index=[("related_weakness", ASCENDING)], name="related_weakness"
                ),
            ],
            "cwe": [
                MongoAddIndex(index=[("id", ASCENDING)], name="id"),
                MongoAddIndex(index=[("name", ASCENDING)], name="name"),
                MongoAddIndex(index=[("status", ASCENDING)], name="status"),
            ],
        }

    def getInfo(self, collection: str):
        return sanitize(self.database["info"].find_one({"db": collection}))

    def create_indexes(self, collection: str = None):
        if collection is not None:
            try:
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, unique=each.unique
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(collection, each.index, name=each.name)
            except KeyError:
                # no specific index given, continue
                self.logger.warning(
                    f"Could not find the requested collection: {collection}, skipping..."
                )
                pass

        else:
            for index in self.iter_indexes():
                self.setIndex(index[0], index[1])

            for collection in self.indexes.keys():
                for each in self.indexes[collection]:
                    if isinstance(each, MongoUniqueIndex):
                        self.setIndex(
                            collection, each.index, name=each.name, unique=each.unique
                        )
                    elif isinstance(each, MongoAddIndex):
                        self.setIndex(collection, each.index, name=each.name)

    def iter_indexes(self):
        for each in self.get_via4_indexes():
            yield each

    def get_via4_indexes(self):
        via4 = self.getInfo("via4")
        result = []
        if via4:
            for index in via4.get("searchables", []):
                result.append(("via4", index))
        return result

    def setIndex(self, col: str, field: str, **kwargs):
        try:
            self.database[col].create_index(field, **kwargs)
            self.logger.info("Success to create index %s on %s" % (field, col))
        except Exception as e:
            self.logger.error("Failed to create index %s on %s: %s" % (col, field, e))



import ast
import collections
import inspect
import json
import logging
from datetime import datetime, timedelta
from itertools import zip_longest
from typing import Any, Iterable
from urllib.parse import urlparse

import pytz
from celery.schedules import schedule, crontab
from redbeat import RedBeatSchedulerEntry
from redbeat.decoder import RedBeatJSONEncoder
from redbeat.schedulers import ensure_conf
from redis import Redis
from redis.exceptions import ConnectionError as RedisConnectionError

from CveXplore.celery_app.cvexplore_daemon import app
from CveXplore.common.config import Configuration
from CveXplore.core.general.utils import timestampTOdatetime
from CveXplore.core.logging.logger_class import AppLogger
from CveXplore.errors.tasks import TaskNotFoundError

logging.setLoggerClass(AppLogger)

TASK_START_MODULE = "CveXplore.celery_app.cvexplore_daemon"
TASK_START_STRING = "CveXplore.celery_app.cvexplore_daemon.crt_"

task_descriptions = {
    k: str(inspect.getdoc(v)).replace("\n", " ")
    for (k, v) in app.tasks.items()
    if k.startswith(TASK_START_STRING)
}


class CveXploreEntry(RedBeatSchedulerEntry):
    """
    CveXploreEntry class is used to create new tasks in the backend; it inherits from RedBeatSchedulerEntry and
    overwrites the save method. By default, the RedBeatSchedulerEntry uses the same redis parameters as the celery
    daemon it facilitates. By overwriting this method the possibility is created to use different redis connection
    parameters then the celery daemon.

    Group:
        taskhandler
    """

    def __init__(
        self,
        name: str = None,
        task: str = None,
        schedule: float | timedelta | schedule | crontab = None,
        args: Any = None,
        kwargs: dict = None,
        enabled: bool = True,
        options: dict = None,
        redis_broker: Redis = None,
        **clsargs,
    ):
        """
        Create a new instance of CveXploreEntry.

        Args:
            name: Slug of the task
            task: Task name as stated in celery daemon definition
            schedule: Schedule of the task
            args: Arguments of the task
            kwargs: Keyword arguments of the task
            enabled: Whether the task is enabled or not
            options: Options of the task
            redis_broker: Redis instance to connect to redis broker
            **clsargs: Class arguments of the task
        """
        super().__init__(
            name=str(name),
            task=task,
            schedule=schedule,
            args=args,
            kwargs=kwargs,
            options=options,
            **clsargs,
        )

        self.enabled = enabled
        ensure_conf(self.app)

        self.redis_broker = redis_broker

    def save(self):
        """
        Save the task to the backend of CveXplore.

        Returns:
            `CveXploreEntry` instance

        """
        definition = {
            "name": self.name,
            "task": self.task,
            "args": self.args,
            "kwargs": self.kwargs,
            "options": self.options,
            "schedule": self.schedule,
            "enabled": self.enabled,
        }
        meta = {
            "last_run_at": self.last_run_at,
        }
        with self.redis_broker.pipeline() as pipe:
            pipe.hset(
                self.key, "definition", json.dumps(definition, cls=RedBeatJSONEncoder)
            )
            pipe.hsetnx(self.key, "meta", json.dumps(meta, cls=RedBeatJSONEncoder))
            pipe.zadd(self.app.redbeat_conf.schedule_key, {self.key: self.score})
            pipe.execute()

        return self


class Task(object):
    """
    The Task class is used to handle all task related operations related to the CveXplore backend.

    Group:
        taskhandler
    """

    def __init__(
        self,
        name: str,
        task: str,
        run: schedule | crontab = None,
        args: Any = None,
        kwargs: dict = None,
        enabled: bool = True,
        last_run_at: datetime = None,
        total_run_count: int = 0,
        next_run_at: datetime = None,
    ):
        """
        Create a new instance of Task

        Args:
            name: Slug of the task
            task: Task name as stated in celery daemon definition
            run: Schedule of the task
            args: Arguments of the task
            kwargs: Keyword arguments of the task
            enabled: Whether the task is enabled or not
            last_run_at: Last run time of the task
            total_run_count: Total run count of the task
            next_run_at: Next run time of the task
        """
        self.name = name
        self.task = (
            task if task.startswith(TASK_START_STRING) else f"{TASK_START_STRING}{task}"
        )
        self.run = run
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.kwargs["task_slug"] = self.name
        self.enabled = enabled

        try:
            self.description = task_descriptions[self.task]
            self.periodic_daemon_task = False
        except KeyError:
            # daemon task
            self.description = ""
            self.periodic_daemon_task = True

        self.last_run_at = last_run_at
        self.total_run_count = total_run_count
        self.next_run_at = next_run_at

        self.task_start_string = TASK_START_STRING

        if isinstance(self.run, schedule):
            self.crontab = False
        else:
            self.crontab = True

        self.config = Configuration

        parsed_url = urlparse(self.config.REDIS_URL)

        self.redis_broker = Redis(
            host=parsed_url.hostname,
            port=parsed_url.port,
            db=int(self.config.CELERY_REDIS_BROKER_DB),
        )

        self.redis_backend = Redis(
            host=parsed_url.hostname,
            port=parsed_url.port,
            db=int(self.config.CELERY_REDIS_BACKEND_DB),
        )

        status = self.redis_backend.get(f"runresult_{self.name}")

        if status is not None:
            if isinstance(status, dict):
                status = self.decode_redis_output(status)
            else:
                status = int(status.decode())

        self.last_run_result = status

        self.logger = logging.getLogger(__name__)

    @property
    def is_enabled(self):
        """
        Property to check if the task is enabled or not.

        Returns:
            Whether the task is enabled or not; True when the task is enabled, False when the task is disabled.
        Group:
            properties
        """
        return self.enabled

    def batcher(self, iterable: Iterable, n: int) -> Iterable:
        """
        Helper function to request chunks of keys from the Redis backend.

        Args:
            iterable:
            n:

        Returns:
            A list of Redis keys
        """
        args = [iter(iterable)] * n
        return zip_longest(*args)

    def decode_redis_output(
        self, src: list | dict | bytes | None
    ) -> list | dict | str | None:
        """
        Helper function to decode output from Redis backend.

        Args:
            src: redis output to decode

        Returns:
            A list, dict or str with decoded output or None if src is None.
        """
        if isinstance(src, list):
            rv = list()
            for key in src:
                rv.append(self.decode_redis_output(key))
            return rv
        elif isinstance(src, dict):
            rv = dict()
            for key in src:
                if not isinstance(key, str):
                    rv[key.decode()] = self.decode_redis_output(src[key])
                else:
                    rv[key] = self.decode_redis_output(src[key])
            return rv
        elif isinstance(src, bytes):
            try:
                return ast.literal_eval(src.decode())
            except Exception:
                return src.decode()
        elif src is None:
            return src
        else:
            raise Exception("type not handled: " + type(src))

    def get_all_task_results(self) -> Iterable[list | dict | str | None]:
        """
        Iterator over all task results in the Redis backend.

        Returns:
            List with Redis task data from the backend
        """
        for keybatch in self.batcher(
            self.redis_backend.scan_iter(f"{self.name}_*"), 500
        ):
            my_pipeline = self.redis_backend.pipeline()
            cleaned_keybatch = [x.decode("utf-8") for x in keybatch if x is not None]
            for key in cleaned_keybatch:
                my_pipeline.hgetall(key)
            data = []
            for d in my_pipeline.execute():
                cleaned_data = self.decode_redis_output(d)
                data.append(cleaned_data)

            yield data

    def get_sorted_task_results(
        self, limit: int, desc: bool = True
    ) -> Iterable[list | dict | str | None]:
        """
        Get task results in the Redis backend.

        Args:
            limit: Limit the amount of results to return.
            desc: If desc is True, sort descending.

        Returns:
            Sorted list of task results.
        """
        task_results = self.redis_backend.zrange(
            f"sortresults_{self.name}", 0, limit, desc=desc, withscores=True
        )
        # fetch the keys
        key_list = [x[0] for x in task_results]

        my_pipeline = self.redis_backend.pipeline()
        cleaned_keybatch = [x.decode("utf-8") for x in key_list if x is not None]
        for key in cleaned_keybatch:
            my_pipeline.hgetall(key)
        data = []
        for d in my_pipeline.execute():
            cleaned_data = self.decode_redis_output(d)
            data.append(cleaned_data)

        return data

    def purge_task_results(self) -> bool:
        """
        Purge task results in the Redis backend.

        Raises:
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task results are purged.
        """
        try:
            count = 0

            for keybatch in self.batcher(
                self.redis_backend.scan_iter(f"{self.name}_*"), 500
            ):
                cleaned_keybatch = [x for x in keybatch if x is not None]
                count += len(cleaned_keybatch)
                self.redis_backend.delete(*cleaned_keybatch)

            self.total_run_count = 0
            self.upsert_task()

            all_run_results = self.redis_backend.keys(f"runresult_{self.name}*")
            if len(all_run_results) != 0:
                self.redis_backend.unlink(*all_run_results)

            self.redis_backend.unlink(f"sortresults_{self.name}")

            entry = self.redis_broker.hgetall("redbeat:{}".format(self.name))

            new_entry = json.loads(entry[b"meta"])

            new_entry["total_run_count"] = 0

            entry[b"meta"] = json.dumps(new_entry).encode()

            self.redis_broker.hset("redbeat:{}".format(self.name), mapping=entry)

            self.logger.info(
                f"Purged {self.name} database entries, deleting: {count} records"
            )

            return True
        except Exception as e:
            self.logger.error(f"Error purging task results: {e}")
            raise

    def upsert_task(self) -> bool:
        """
        Method to create or update a scheduled interval task

        Raises:
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task is saved.
        """
        try:
            entry = CveXploreEntry(
                name=self.name,
                task=self.task,
                schedule=self.run,
                args=self.args,
                kwargs=self.kwargs,
                enabled=self.enabled,
                app=app,
                redis_broker=self.redis_broker,
            )
            entry.save()

            return True
        except Exception as e:
            self.logger.exception(f"Failed to save the task!! Error: {e}")
            raise

    def delete_task(self) -> bool:
        """
        Method to delete a task from the database

        Raises:
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task is deleted.
        """
        try:
            self.purge_task_results()
            self.redis_backend.delete(f"runresult_{self.name}")
            self.redis_broker.zrem("redbeat::schedule", "redbeat:{}".format(self.name))
            self.redis_broker.delete("redbeat:{}".format(self.name))

            return True
        except Exception:
            self.logger.exception("Failed to delete the task!!")
            raise

    def enable(self) -> bool:
        """
        Method to enable the task

        Returns:
            True if the task is enabled, False otherwise.
        """
        return self._toggle_task(1)

    def disable(self) -> bool:
        """
        Method to disable the task.

        Returns:
            True if the task is disabled, False otherwise.
        """
        return self._toggle_task(0)

    def _toggle_task(self, value: int) -> bool:
        """
        method to enable (1) or disable (0) a task

        Raises:
            Exception: If an uncaught exception is raised.

        Returns:
            True if the toggle succeeds, False otherwise.
        """

        if value == 0:
            set_value = False
        elif value == 1:
            set_value = True

        try:
            entry = self.redis_broker.hgetall("redbeat:{}".format(self.name))

            new_entry = json.loads(entry[b"definition"])

            new_entry["enabled"] = set_value
            self.enabled = set_value

            entry[b"definition"] = json.dumps(new_entry).encode()

            self.redis_broker.hset("redbeat:{}".format(self.name), mapping=entry)

            return True
        except Exception:
            self.logger.exception("Failed to toggle the task!!")
            raise

    def to_dict(self) -> dict:
        """
        Method to convert the `Task` result to a dictionary.

        Returns:
            Serialized `Task` data
        """
        my_dict = {
            k: v
            for (k, v) in self.__dict__.items()
            if not k.startswith("_")
            and k
            not in {
                "logger",
                "enable",
                "disable",
                "delete_task",
                "upsert_task",
                "config",
                "redis_broker",
                "redis_backend",
            }
        }

        if not self.crontab:
            my_dict["seconds"] = int(my_dict["run"].seconds)
            my_dict["run"] = my_dict["run"].human_seconds

        return my_dict

    def to_data(self):
        """
        Method to convert the `Task` result to a dictionary.

        Returns:
            Serialized `Task` data
        """
        my_dict = {
            k: v
            for (k, v) in self.__dict__.items()
            if not k.startswith("_")
            and k
            not in {
                "logger",
                "enable",
                "disable",
                "delete_task",
                "upsert_task",
                "config",
                "redis_broker",
                "redis_backend",
            }
        }

        if not self.crontab:
            my_dict["seconds"] = int(my_dict["run"].seconds)
            my_dict["run"] = my_dict["run"].human_seconds

        if self.crontab:
            my_dict.pop("run")
            my_dict["minute"] = self.run._orig_minute
            my_dict["hour"] = self.run._orig_hour
            my_dict["dow"] = self.run._orig_day_of_week
            my_dict["dom"] = self.run._orig_day_of_month
            my_dict["moy"] = self.run._orig_month_of_year

        if len(self.args) != 0:
            args_str = ""
            for each in self.args:
                if isinstance(each, str):
                    args_str += f"{each}\n"
                elif isinstance(each, list):
                    args_str += ",".join(each) + "\n"
                elif isinstance(each, dict):
                    args_str += (
                        ",".join(
                            [
                                f"{':'.join(map(str,v))}"
                                for i, v in enumerate(list(each.items()))
                            ]
                        )
                        + "\n"
                    )
            my_dict["args"] = args_str

        if len(self.kwargs) != 0:
            kwargs_str = ""
            for the_key, the_val in self.kwargs.items():
                if isinstance(self.kwargs[the_key], str):
                    kwargs_str += f"{the_key}={the_val}\n"
                elif isinstance(self.kwargs[the_key], list):
                    kwargs_str += f"{the_key}={','.join(the_val)}\n"
                elif isinstance(self.kwargs[the_key], dict):
                    d = ",".join(
                        [
                            f"{':'.join(map(str, v))}"
                            for i, v in enumerate(list(the_val.items()))
                        ]
                    )
                    kwargs_str += f"{the_key}={d}\n"
            my_dict["kwargs"] = kwargs_str

        return my_dict

    def __repr__(self):
        """
        String representation of the object.

        Returns:
            String representation of the object.
        """
        return f"<< Task: {self.name} >>"


class TaskData(object):
    """
    Class to hold task data.

    Group:
        taskhandler
    """

    def __init__(self, entry_data: dict):
        """
        Method to initialize the `TaskData` object.

        Args:
            entry_data: Dictionary of task data.
        """
        self.entry_data = entry_data

        for k, v in self.entry_data[b"definition"].items():
            self.__setattr__(k, v)

        if hasattr(self, "schedule"):
            if self.schedule["__type__"] == "interval":
                self.run = schedule(
                    run_every=self.schedule["every"],
                    relative=self.schedule["relative"],
                )
            if self.schedule["__type__"] == "crontab":
                task_data = self.schedule
                task_data.pop("__type__")
                self.run = crontab(**task_data)

        if "total_run_count" in self.entry_data[b"meta"]:
            self.total_run_count = self.entry_data[b"meta"]["total_run_count"]
        else:
            self.total_run_count = 0

        self.last_run_at = datetime(
            year=self.entry_data[b"meta"]["last_run_at"]["year"],
            month=self.entry_data[b"meta"]["last_run_at"]["month"],
            day=self.entry_data[b"meta"]["last_run_at"]["day"],
            hour=self.entry_data[b"meta"]["last_run_at"]["hour"],
            minute=self.entry_data[b"meta"]["last_run_at"]["minute"],
            second=self.entry_data[b"meta"]["last_run_at"]["second"],
            microsecond=self.entry_data[b"meta"]["last_run_at"]["microsecond"],
            tzinfo=pytz.utc,
        )

    def to_dict(self) -> dict:
        """
        Method to convert the `TaskData` result to a dictionary.

        Returns:
            Serialized `TaskData` data
        """
        return {
            "name": self.name,
            "task": self.task,
            "run": self.run,
            "args": self.args,
            "kwargs": self.kwargs,
            "enabled": self.enabled,
            "last_run_at": self.last_run_at,
            "total_run_count": self.total_run_count,
        }

    def __repr__(self):
        """
        String representation of the object.

        Returns:
            String representation of the object.
        """
        return f"<< TaskData: {self.name} >>"


class TaskHandler(object):
    """
    Class that is the main handler towards the backend; this class is used by the package and CLI for
    all CRUD operations on the backend of CveXplore.

    Group:
        taskhandler
    """

    def __init__(self):
        """
        Method to initialize the `TaskHandler` object.
        """
        self.config = Configuration

        parsed_url = urlparse(self.config.REDIS_URL)

        self.redis = Redis(
            host=parsed_url.hostname,
            port=parsed_url.port,
            db=int(self.config.CELERY_REDIS_BROKER_DB),
        )

        self.logger = logging.getLogger(__name__)

    @staticmethod
    def show_available_tasks() -> collections.OrderedDict:
        """
        Show available tasks. Sorted by key.

        Returns:
            An ordered dictionary of the available tasks.
        """

        ordered_dict = collections.OrderedDict()

        sorted_tasks = sorted(task_descriptions.keys())

        for each in sorted_tasks:
            ordered_dict[each] = task_descriptions[each]

        return ordered_dict

    def create_task_by_number(
        self,
        task_number: int,
        task_slug: str,
        task_interval: int = None,
        task_crontab: dict = None,
    ) -> bool:
        """
        Create a task by its number. The number should correspond to the indexnumber + 1 in the result of a call to
        `TaskHandler.show_available_tasks()` or should correspond to the number in the ID column of the
        equivalent cli command. Either `task_interval` or `task_crontab` must be specified.

        Examples:

            >>> th = TaskHandler()
            >>> th.create_task_by_number(1, task_slug="test", task_interval=10)
            True

            >>> th = TaskHandler()
            >>> th.create_task_by_number(2, task_slug="test_2", task_crontab={"minute": "*/5"})
            True

        Args:
            task_number: The number of the task.
            task_slug: The slug that you want to use for the task.
            task_interval: The interval (in seconds) that you want to run the task.
            task_crontab: The crontab that you want to use for the task.

        Raises:
            ValueError: if both `task_interval` and `task_crontab` are not specified.
            TaskNotFoundError: If the specified `task_number` is not found.
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task was successfully created.
        """
        try:
            all_tasks = self.show_available_tasks()
            task_name = list(all_tasks.keys())[task_number - 1]

            return self._schedule_task(
                task_slug=task_slug,
                task_name=task_name,
                task_interval=task_interval,
                task_crontab=task_crontab,
            )
        except IndexError:
            raise TaskNotFoundError
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def show_scheduled_tasks(self) -> list[Task]:
        """
        Show all scheduled tasks. Sorted by `Task.name`. A scheduled task is a task that is inserted into the redis
        beat queue and is executed according to the `task_interval` or `task_crontab` variables.

        Examples:

            >>> th = TaskHandler()
            >>> th.show_scheduled_tasks()
            [<< Task: test >>, << Task: test_2 >>]

        Raises:
            RedisConnectionError: If connection to Redis could not be established.
            Exception: If an uncaught exception is raised.

        Returns:
            A list of scheduled tasks.
        """
        try:
            tasks = self.redis.zrange("redbeat::schedule", 0, -1, withscores=True)

            ret_list = []

            for task_details in tasks:
                task, next_run = task_details

                entry = self.redis.hgetall(task)

                for each in entry:
                    entry[each] = json.loads(entry[each])

                task_data = TaskData(entry)

                ret_list.append(
                    Task(
                        **task_data.to_dict(),
                        next_run_at=timestampTOdatetime(int(next_run)),
                    )
                )

            ret_list = sorted(ret_list, key=lambda x: x.name.lower())

            return ret_list
        except RedisConnectionError as err:
            self.logger.error(f"Redis connection error; {err}")
            raise
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def delete_scheduled_task(self, task_id: int) -> bool:
        """
        Delete a scheduled task by number. The number should correspond to the indexnumber + 1 in the result of a
        call to `TaskHandler.show_scheduled_tasks()` or should correspond to the number in the ID column of the
        equivalent cli command.

        Args:
            task_id: The number of the task.

        Raises:
            TaskNotFoundError: If the specified `task_number` is not found.
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task was successfully deleted.
        """
        try:
            all_tasks = self.show_scheduled_tasks()
            the_task = all_tasks[task_id - 1]
            return the_task.delete_task()
        except IndexError:
            raise TaskNotFoundError
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def toggle_scheduled_task(self, task_id: int) -> bool:
        """
        Toggle a scheduled task by number between an enabled and disabled state.

        Args:
            task_id: The number of the task.

        Raises:
            TaskNotFoundError: If the specified `task_number` is not found.
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task was successfully toggled.
        """
        try:
            all_tasks = self.show_scheduled_tasks()
            the_task = all_tasks[task_id - 1]
            if the_task.enabled:
                return the_task.disable()
            else:
                return the_task.enable()
        except IndexError:
            raise TaskNotFoundError
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def purge_scheduled_task(self, task_id: int) -> bool:
        """
        Purge the results from a given task. The number should correspond to the indexnumber + 1 in the result of a
        call to `TaskHandler.show_scheduled_tasks()` or should correspond to the number in the ID column of the
        equivalent cli command.

        Args:
            task_id: The number of the task.

        Raises:
            TaskNotFoundError: If the specified `task_number` is not found.
            Exception: If an uncaught exception is raised.

        Returns:
            True if the task was successfully purged.
        """
        try:
            all_tasks = self.show_scheduled_tasks()
            the_task = all_tasks[task_id - 1]
            return the_task.purge_task_results()
        except IndexError:
            raise TaskNotFoundError
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def get_scheduled_tasks_results(self, task_id: int, limit: int = 10) -> list[dict]:
        """
        Method to retrieve the results of a scheduled task by number. The number should correspond to the
        indexnumber + 1 in the result of a call to `TaskHandler.show_scheduled_tasks()` or should correspond to the
        number in the ID column of the equivalent cli command.

        Args:
            task_id: The number of the task.
            limit: The maximum number of results to retrieve.

        Raises:
            TaskNotFoundError: If the specified `task_number` is not found.
            Exception: If an uncaught exception is raised.

        Returns:
            A list with dictionaries containing the results of the scheduled task.
        """
        try:
            all_tasks = self.show_scheduled_tasks()
            the_task = all_tasks[task_id - 1]
            task_results = the_task.get_sorted_task_results(limit=limit)
            return sorted(task_results, key=lambda x: x["inserted"], reverse=True)
        except IndexError:
            raise TaskNotFoundError
        except Exception as err:
            self.logger.error(
                f"Uncaught exception while retrieving scheduled tasks: {err}"
            )
            raise

    def get_scheduled_task_by_name(self, task_name: str) -> Task:
        """
        Method to retrieve the task parameters

        Args:
            task_name: The task slug

        Raises:
            TaskNotFoundError: If the specified `task_number` is not found.

        Returns:
            A `Task` instance.
        """
        try:
            task = [
                x
                for x in self.show_scheduled_tasks()
                if x.name == task_name
                or x.name.startswith(f"{TASK_START_STRING}{task_name}")
            ][0]

            return task
        except IndexError:
            raise TaskNotFoundError

    def _schedule_task(
        self,
        task_slug: str,
        task_name: str,
        task_interval: int = None,
        task_crontab: dict = None,
        task_args: list = None,
        task_kwargs: dict = None,
        task_enabled: bool = True,
    ) -> bool:
        """
        Method to schedule a task; used by `TaskHandler.create_task_by_number()`.

        Args:
            task_slug: The slug that you want to use for the task
            task_name: The name of the task
            task_interval: The interval in seconds that the task should be scheduled
            task_crontab: The crontab dictionary that will be used to schedule a task
            task_args: The arguments that will be passed to the task
            task_kwargs: The keyword arguments that will be passed to the task
            task_enabled: Whether the task should be enabled or not.

        Raises:
            ValueError: if both `task_interval` and `task_crontab` are not specified.
            Exception: If an uncaught exception is raised.

        Returns:

        """
        if task_interval is not None:
            job_run = schedule(run_every=task_interval)
        elif task_crontab is not None:
            job_run = crontab(**task_crontab)
        else:
            raise ValueError("Either task_interval or task_crontab must be provided")

        try:
            my_task = Task(
                name=task_slug,
                task=task_name,
                run=job_run,
                args=task_args,
                kwargs=task_kwargs,
                enabled=task_enabled,
            )
            return my_task.upsert_task()
        except Exception as e:
            self.logger.error(f"Error creating task: {e}")
            raise

    def __repr__(self):
        """
        String representation of the object.

        Returns:
            String representation of the object.
        """
        return "<< TaskHandler >>"



import logging

import click
import click_completion.core

from CveXplore.cli_cmds.capec_cmds import commands as group6
from CveXplore.cli_cmds.cpe_cmds import commands as group5
from CveXplore.cli_cmds.cve_cmds import commands as group2
from CveXplore.cli_cmds.cwe_cmds import commands as group7
from CveXplore.cli_cmds.db_cmds import commands as group4
from CveXplore.cli_cmds.find_cmds import commands as group1
from CveXplore.cli_cmds.stats_cmds import commands as group3
from CveXplore.cli_cmds.tasks_cmds import commands as group8
from CveXplore.main import CveXplore

click_completion.init()

logging.getLogger("dicttoxml").setLevel("ERROR")


@click.group(invoke_without_command=True)
@click.option("-v", "--version", is_flag=True, help="Show the current version and exit")
@click.option(
    "-w", "--where", is_flag=True, help="Show the path where CveXplore is installed"
)
@click.pass_context
def main(ctx, version, where):
    ctx.obj = {"data_source": CveXplore()}
    if version:
        click.echo(ctx.obj["data_source"].version)
        exit(0)
    elif where:
        click.echo(ctx.obj["data_source"].where())
        exit(0)
    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))


main.add_command(group1.find_cmd)
main.add_command(group2.cve_cmd)
main.add_command(group3.stats_cmd)
main.add_command(group4.db_cmd)
main.add_command(group5.cpe_cmd)
main.add_command(group6.capec_cmd)
main.add_command(group7.cwe_cmd)
main.add_command(group8.tasks_cmd)



import os
from itertools import chain

from setuptools import setup, find_packages

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))


# Get static version from CveXplore/VERSION
def _version():
    version_file = os.path.join(HERE, "CveXplore/VERSION")

    with open(version_file, "r") as fdsec:
        tag = fdsec.read()
    return tag


VERSION = __version__ = _version()

# The text of the README file
with open(os.path.join(HERE, "README.rst")) as fid:
    README = fid.read().split("##INCLUDE_MARKER##")[1]


# -*- Extras -*-
MODULES = {
    "all-db",
    "celery",
    "docs",
    "kafka",
    "mongodb",
    "mysql",
    "postgres",
    "redis",
    "sqlalchemy",
    "sqllite",
}


# -*- Requirements -*-
def _strip_comments(l):
    return l.split("#", 1)[0].strip()


def _pip_requirement(req):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*path.split("/"))
    return [req]


def _reqs(*f):
    if len(f) == 2:
        if os.getcwd().endswith("modules") and f[0] == "modules":
            f = [f[1]]
        if not os.getcwd().endswith("modules") and f[0] == ".":
            f = ("modules", f[1])
    return [
        _pip_requirement(r)
        for r in (
            _strip_comments(l)
            for l in open(os.path.join(os.getcwd(), "requirements", *f)).readlines()
        )
        if r
    ]


def reqs(*f):
    """Parse requirement file.

    Example:
        reqs('default.txt')          # requirements/default.txt
        reqs('modules', 'redis.txt')  # requirements/modules/redis.txt
        reqs('.', 'loggers.txt')  # requirements/modules/loggers.txt -> this is a reference in a requirements file
        like -r ./loggers.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p):
    """Parse requirement in the requirements/modules/ directory."""
    return reqs("modules", *p)


def install_requires():
    """Get list of requirements required for installation."""
    return reqs("default.txt")


def extras_require():
    """Get map of all extra requirements."""
    module_requirements = {x: extras(x + ".txt") for x in MODULES}

    # add an 'all' value to install all requirements for all modules
    module_requirements["all"] = list(set(chain(*module_requirements.values())))

    return module_requirements


setup(
    name="CveXplore",
    version=VERSION,
    packages=find_packages(exclude=("tests", "docs")),
    url="https://github.com/cve-search/CveXplore",
    license="GNU General Public License v3.0",
    author="Paul Tikken",
    author_email="paul.tikken@gmail.com",
    home_page="https://github.com/cve-search/CveXplore",
    project_urls={
        "Documentation": "https://cve-search.github.io/CveXplore/",
        "Issues": "https://github.com/cve-search/CveXplore/issues",
    },
    description="Package for interacting with cve-search",
    long_description=README,
    long_description_content_type="text/x-rst",
    package_data={
        "CveXplore": [
            "LICENSE",
            "VERSION",
            ".schema_version",
            "common/.sources.ini",
            "common/.env_example",
            "alembic/*",
            "alembic/**/*",
            "alembic.ini",
            "backend/*",
            "backend/**/*",
            "backend/**/**/*",
        ]
    },
    entry_points="""
            [console_scripts]
            cvexplore=CveXplore.cli:main
        """,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=install_requires(),
    extras_require=extras_require(),
)



import click
import pymongo

from CveXplore.cli_cmds.cli_utils.utils import printer
from CveXplore.cli_cmds.mutex_options.mutex import Mutex


@click.group("cpe", invoke_without_command=True, help="Query for cpe specific data")
@click.pass_context
def cpe_cmd(ctx):
    if ctx.invoked_subcommand is None:
        click.echo(cpe_cmd.get_help(ctx))


@cpe_cmd.group(
    "search",
    invoke_without_command=True,
    help="Search for cpe entries by name, vendor or title. Results are sorted ASCENDING by default",
)
@click.option(
    "-n",
    "--name",
    default=False,
    is_flag=True,
    help="Search by name",
    cls=Mutex,
    not_required_if=["field", "cpe", "title", "vendor", "field_list"],
)
@click.option(
    "-t",
    "--title",
    default=True,
    is_flag=True,
    help="Search by title (default)",
    cls=Mutex,
    not_required_if=["field", "cpe", "name", "vendor", "field_list"],
)
@click.option(
    "-v",
    "--vendor",
    default=False,
    is_flag=True,
    help="Search by vendor",
    cls=Mutex,
    not_required_if=["field", "cpe", "name", "title", "field_list"],
)
@click.option(
    "-c",
    "--cpe",
    help="Search for CPE's (could be multiple) by id",
    multiple=True,
    cls=Mutex,
    not_required_if=["name", "title", "vendor"],
)
@click.option(
    "-f",
    "--field",
    help="Field to return (could be multiple)",
    multiple=True,
    cls=Mutex,
    not_required_if=["field_list", "name", "title", "vendor"],
)
@click.option(
    "-fl",
    "--field_list",
    help="Return a field list for this collection",
    is_flag=True,
    cls=Mutex,
    not_required_if=["field", "name", "title", "vendor"],
)
@click.option(
    "-m",
    "--match",
    help="Use match for searching",
    cls=Mutex,
    not_required_if=["regex"],
)
@click.option(
    "-r",
    "--regex",
    help="Use regex for searching",
    cls=Mutex,
    not_required_if=["match"],
)
@click.option("-d", "--deprecated", is_flag=True, help="Filter deprecated cpe's")
@click.option("-rc", "--related_cve", is_flag=True, help="Add related CVE's")
@click.option(
    "-p",
    "--product_search",
    default=False,
    is_flag=True,
    help="Let CveXplore search for vulnerable products and not for vulnerable configurations",
)
@click.option("-l", "--limit", default=10, help="Search limit")
@click.option("-s", "--sort", is_flag=True, help="Sort DESCENDING")
@click.option(
    "-o",
    "--output",
    default="json",
    help="Set the desired output format (defaults to json)",
    type=click.Choice(["json", "csv", "xml", "html"], case_sensitive=False),
)
@click.pass_context
def search_cmd(
    ctx,
    name,
    title,
    vendor,
    cpe,
    field,
    field_list,
    match,
    regex,
    deprecated,
    related_cve,
    product_search,
    limit,
    sort,
    output,
):
    if not name and not vendor and title:
        search_by = "title"
    elif name and not vendor and title:
        search_by = "cpeName"
    else:
        search_by = "vendor"

    if not sort:
        sorting = pymongo.ASCENDING
    else:
        sorting = pymongo.DESCENDING

    if regex:
        if deprecated:
            ret_list = getattr(ctx.obj["data_source"], "cpe").search_active_cpes(
                search_by, regex, limit, sorting
            )
        else:
            ret_list = (
                getattr(getattr(ctx.obj["data_source"], "cpe"), search_by)
                .search(regex)
                .limit(limit)
                .sort(search_by, sorting)
            )
    elif match:
        if deprecated:
            ret_list = getattr(ctx.obj["data_source"], "cpe").find_active_cpes(
                search_by, match, limit, sorting
            )
        else:
            ret_list = (
                getattr(getattr(ctx.obj["data_source"], "cpe"), search_by)
                .find(match)
                .limit(limit)
                .sort(search_by, sorting)
            )
    elif cpe and not field_list:
        ret_list = getattr(ctx.obj["data_source"], "cpe").mget_by_id(*cpe)
    elif cpe and field_list:
        ret_list = getattr(ctx.obj["data_source"], "cpe").field_list(*cpe)
    else:
        click.echo(search_cmd.get_help(ctx))
        return

    if related_cve:
        result = [result.to_cve_summary(product_search) for result in ret_list]
    elif cpe and not field_list:
        result = [result.to_dict(*field) for result in ret_list]
    elif field_list:
        result = [result for result in ret_list]
    else:
        result = [result.to_dict() for result in ret_list]

    if ctx.invoked_subcommand is None:
        printer(input_data=result, output=output)
    else:
        ctx.obj["RESULT"] = result



import json
import os

from pymongo.collection import Collection

from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase
from CveXplore.database.connection.database_connection import DatabaseConnection
from CveXplore.objects.cvexplore_object import CveXploreObject


class DatasourceConnection(CveXploreObject):
    """
    The DatasourceConnection class handles the connection to the data source and is the base class for the database
    objects and generic database functions

    Group:
        common
    """

    def __init__(self, collection: str):
        """
        Create a DatasourceConnection object

        Args:
            collection: The name of the data source collection

        """
        super().__init__()
        self._collection = collection

    @property
    def datasource_connection(self) -> DatabaseConnectionBase:
        """
        Property to access the datasource connection

        Group:
            properties

        """
        # hack for documentation building
        if json.loads(os.getenv("DOC_BUILD"))["DOC_BUILD"] == "YES":
            return DatabaseConnection(
                database_type="dummy",
                database_init_parameters={},
            ).database_connection
        else:
            return DatabaseConnection(
                database_type=self.config.DATASOURCE_TYPE,
                database_init_parameters=self.config.DATASOURCE_CONNECTION_DETAILS,
            ).database_connection

    @property
    def datasource_collection_connection(self) -> Collection:
        """
        Property to access the datasource collection connection

        Group:
            properties

        """
        return getattr(self.datasource_connection, f"store_{self.collection}")

    @property
    def collection(self) -> str:
        """
        Property to access the collection

        Group:
            properties

        """
        return self._collection

    def to_dict(self, *print_keys: str) -> dict:
        """
        Method to convert the entire object to a dictionary

        Args:
            print_keys: Keys to limit the output dictionary

        Returns:
            A dictionary of the requested keys

        """
        if len(print_keys) != 0:
            full_dict = {
                k: v
                for (k, v) in self.__dict__.items()
                if not k.startswith("_") and k in print_keys
            }
        else:
            full_dict = {
                k: v for (k, v) in self.__dict__.items() if not k.startswith("_")
            }

        return full_dict

    def __eq__(self, other: DatabaseConnection) -> bool:
        return self.__dict__ == other.__dict__

    def __ne__(self, other: DatabaseConnection) -> bool:
        return self.__dict__ != other.__dict__



from CveXplore.database.helpers.generic_db import GenericDatabaseFieldsFunctions

class CvesDatabaseFunctions:
    id: GenericDatabaseFieldsFunctions
    cvss: GenericDatabaseFieldsFunctions
    cvss3: GenericDatabaseFieldsFunctions
    summary: GenericDatabaseFieldsFunctions
    vendors: GenericDatabaseFieldsFunctions
    products: GenericDatabaseFieldsFunctions
    lastModified: GenericDatabaseFieldsFunctions
    modified: GenericDatabaseFieldsFunctions
    published: GenericDatabaseFieldsFunctions
    status: GenericDatabaseFieldsFunctions
    assigner: GenericDatabaseFieldsFunctions
    cwe: GenericDatabaseFieldsFunctions
    epss: GenericDatabaseFieldsFunctions

    def __init__(self, collection: str): ...

class CpeDatabaseFunctions:
    id: GenericDatabaseFieldsFunctions
    title: GenericDatabaseFieldsFunctions
    cpeName: GenericDatabaseFieldsFunctions
    vendor: GenericDatabaseFieldsFunctions
    product: GenericDatabaseFieldsFunctions
    stem: GenericDatabaseFieldsFunctions

    def __init__(self, collection: str): ...

class CapecDatabaseFunctions:
    id: GenericDatabaseFieldsFunctions
    name: GenericDatabaseFieldsFunctions
    summary: GenericDatabaseFieldsFunctions
    prerequisites: GenericDatabaseFieldsFunctions
    solutions: GenericDatabaseFieldsFunctions
    loa: GenericDatabaseFieldsFunctions
    typical_severity: GenericDatabaseFieldsFunctions

    def __init__(self, collection: str): ...

class CWEDatabaseFunctions:
    id: GenericDatabaseFieldsFunctions
    name: GenericDatabaseFieldsFunctions
    status: GenericDatabaseFieldsFunctions
    description: GenericDatabaseFieldsFunctions

    def __init__(self, collection: str): ...



"""Initial full model setup

Revision ID: ecb1788b7e08
Revises: 53df0e286532
Create Date: 2023-12-21 06:39:09.346006

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "ecb1788b7e08"
down_revision: Union[str, None] = "53df0e286532"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "cpe",
        sa.Column("_id", sa.BigInteger(), nullable=False),
        sa.Column("id", sa.String(length=50), nullable=True),
        sa.Column("cpeName", sa.String(length=50), nullable=True),
        sa.Column("cpeNameId", sa.String(length=50), nullable=True),
        sa.Column("created", sa.DateTime(), nullable=True),
        sa.Column("deprecated", sa.Boolean(), nullable=True),
        sa.Column("deprecatedBy", sa.String(length=50), nullable=True),
        sa.Column("lastModified", sa.DateTime(), nullable=True),
        sa.Column("padded_version", sa.String(length=50), nullable=True),
        sa.Column("product", sa.String(length=50), nullable=True),
        sa.Column("stem", sa.String(length=50), nullable=True),
        sa.Column("title", sa.String(length=150), nullable=True),
        sa.Column("vendor", sa.String(length=50), nullable=True),
        sa.Column("version", sa.String(length=50), nullable=True),
        sa.PrimaryKeyConstraint("_id"),
    )
    op.create_index(op.f("ix_cpe__id"), "cpe", ["_id"], unique=True)
    op.create_index(op.f("ix_cpe_cpeName"), "cpe", ["cpeName"], unique=False)
    op.create_index(op.f("ix_cpe_deprecated"), "cpe", ["deprecated"], unique=False)
    op.create_index(op.f("ix_cpe_id"), "cpe", ["id"], unique=True)
    op.create_index(op.f("ix_cpe_lastModified"), "cpe", ["lastModified"], unique=False)
    op.create_index(
        op.f("ix_cpe_padded_version"), "cpe", ["padded_version"], unique=False
    )
    op.create_index(op.f("ix_cpe_product"), "cpe", ["product"], unique=False)
    op.create_index(op.f("ix_cpe_stem"), "cpe", ["stem"], unique=False)
    op.create_index(op.f("ix_cpe_title"), "cpe", ["title"], unique=False)
    op.create_index(op.f("ix_cpe_vendor"), "cpe", ["vendor"], unique=False)
    op.create_table(
        "cpeother",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_cpeother_id"), "cpeother", ["id"], unique=True)
    op.create_table(
        "cves",
        sa.Column("_id", sa.BigInteger(), nullable=False),
        sa.Column("id", sa.String(length=50), nullable=True),
        sa.Column("access", sa.JSON(), nullable=True),
        sa.Column("assigner", sa.String(length=50), nullable=True),
        sa.Column("cvss", sa.Float(), nullable=True),
        sa.Column("cvss3", sa.Float(), nullable=True),
        sa.Column("cvssSource", sa.String(length=50), nullable=True),
        sa.Column("cvssTime", sa.DateTime(), nullable=True),
        sa.Column("cvssVector", sa.String(length=100), nullable=True),
        sa.Column("cwe", sa.String(length=50), nullable=True),
        sa.Column("epss", sa.Float(), nullable=True),
        sa.Column("epssMetric", sa.JSON(), nullable=True),
        sa.Column("exploitabilityScore", sa.Float(), nullable=True),
        sa.Column("impact", sa.JSON(), nullable=True),
        sa.Column("impactScore", sa.Float(), nullable=True),
        sa.Column("lastModified", sa.DateTime(), nullable=True),
        sa.Column("modified", sa.DateTime(), nullable=True),
        sa.Column("products", sa.JSON(), nullable=True),
        sa.Column("published", sa.DateTime(), nullable=True),
        sa.Column("references", sa.JSON(), nullable=True),
        sa.Column("status", sa.String(length=25), nullable=True),
        sa.Column("summary", sa.Text(), nullable=True),
        sa.Column("vendors", sa.JSON(), nullable=True),
        sa.Column("vulnerable_configuration", sa.JSON(), nullable=True),
        sa.Column("vulnerable_configuration_cpe_2_2", sa.JSON(), nullable=True),
        sa.Column("vulnerable_configuration_stems", sa.JSON(), nullable=True),
        sa.Column("vulnerable_product", sa.JSON(), nullable=True),
        sa.Column("vulnerable_product_stems", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("_id"),
    )
    op.create_index(op.f("ix_cves__id"), "cves", ["_id"], unique=True)
    op.create_index(op.f("ix_cves_assigner"), "cves", ["assigner"], unique=False)
    op.create_index(op.f("ix_cves_cvss"), "cves", ["cvss"], unique=False)
    op.create_index(op.f("ix_cves_cvss3"), "cves", ["cvss3"], unique=False)
    op.create_index(op.f("ix_cves_cwe"), "cves", ["cwe"], unique=False)
    op.create_index(op.f("ix_cves_epss"), "cves", ["epss"], unique=False)
    op.create_index(op.f("ix_cves_id"), "cves", ["id"], unique=True)
    op.create_index(
        op.f("ix_cves_lastModified"), "cves", ["lastModified"], unique=False
    )
    op.create_index(op.f("ix_cves_modified"), "cves", ["modified"], unique=False)
    op.create_index(op.f("ix_cves_published"), "cves", ["published"], unique=False)
    op.create_index(op.f("ix_cves_status"), "cves", ["status"], unique=False)
    op.create_table(
        "cwe",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("name", sa.String(length=250), nullable=True),
        sa.Column("status", sa.String(length=25), nullable=True),
        sa.Column("weaknessabs", sa.String(length=25), nullable=True),
        sa.Column("related_weaknesses", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_cwe_id"), "cwe", ["id"], unique=True)
    op.create_index(op.f("ix_cwe_name"), "cwe", ["name"], unique=False)
    op.create_index(op.f("ix_cwe_status"), "cwe", ["status"], unique=False)
    op.create_table(
        "mgmt_blacklist",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_mgmt_blacklist_id"), "mgmt_blacklist", ["id"], unique=True)
    op.create_table(
        "mgmt_whitelist",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_mgmt_whitelist_id"), "mgmt_whitelist", ["id"], unique=True)
    op.create_table(
        "via4",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.String(length=50), nullable=True),
        sa.Column("db", sa.String(length=25), nullable=True),
        sa.Column("searchables", sa.JSON(), nullable=True),
        sa.Column("sources", sa.JSON(), nullable=True),
        sa.Column("msbulletin", sa.JSON(), nullable=True),
        sa.Column("redhat", sa.JSON(), nullable=True),
        sa.Column("refmap", sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint("_id"),
    )
    op.create_index(op.f("ix_via4__id"), "via4", ["_id"], unique=True)
    op.create_index(op.f("ix_via4_id"), "via4", ["id"], unique=False)
    op.create_index(
        op.f("ix_capec_typical_severity"), "capec", ["typical_severity"], unique=False
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_capec_typical_severity"), table_name="capec")
    op.drop_index(op.f("ix_via4_id"), table_name="via4")
    op.drop_index(op.f("ix_via4__id"), table_name="via4")
    op.drop_table("via4")
    op.drop_index(op.f("ix_mgmt_whitelist_id"), table_name="mgmt_whitelist")
    op.drop_table("mgmt_whitelist")
    op.drop_index(op.f("ix_mgmt_blacklist_id"), table_name="mgmt_blacklist")
    op.drop_table("mgmt_blacklist")
    op.drop_index(op.f("ix_cwe_status"), table_name="cwe")
    op.drop_index(op.f("ix_cwe_name"), table_name="cwe")
    op.drop_index(op.f("ix_cwe_id"), table_name="cwe")
    op.drop_table("cwe")
    op.drop_index(op.f("ix_cves_status"), table_name="cves")
    op.drop_index(op.f("ix_cves_published"), table_name="cves")
    op.drop_index(op.f("ix_cves_modified"), table_name="cves")
    op.drop_index(op.f("ix_cves_lastModified"), table_name="cves")
    op.drop_index(op.f("ix_cves_id"), table_name="cves")
    op.drop_index(op.f("ix_cves_epss"), table_name="cves")
    op.drop_index(op.f("ix_cves_cwe"), table_name="cves")
    op.drop_index(op.f("ix_cves_cvss3"), table_name="cves")
    op.drop_index(op.f("ix_cves_cvss"), table_name="cves")
    op.drop_index(op.f("ix_cves_assigner"), table_name="cves")
    op.drop_index(op.f("ix_cves__id"), table_name="cves")
    op.drop_table("cves")
    op.drop_index(op.f("ix_cpeother_id"), table_name="cpeother")
    op.drop_table("cpeother")
    op.drop_index(op.f("ix_cpe_vendor"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_title"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_stem"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_product"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_padded_version"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_lastModified"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_id"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_deprecated"), table_name="cpe")
    op.drop_index(op.f("ix_cpe_cpeName"), table_name="cpe")
    op.drop_index(op.f("ix_cpe__id"), table_name="cpe")
    op.drop_table("cpe")
    # ### end Alembic commands ###



import logging

import ijson

from CveXplore.core.database_maintenance.update_base_class import UpdateBaseClass
from CveXplore.core.logging.logger_class import AppLogger

logging.setLoggerClass(AppLogger)


class IJSONHandler(UpdateBaseClass):
    def __init__(self):
        super().__init__(__name__)

    def fetch(self, filename: str, prefix: str):
        x = 0
        with open(filename, "rb") as input_file:
            for item in ijson.items(input_file, prefix):
                yield item
                x += 1

        self.logger.debug(
            f"Processed {x} items from file: {filename}, using prefix: {prefix}"
        )
